/***************************************************************************
 * CVSID: $Id: libmnt.c,v 1.23 2004/05/26 19:12:39 stefanb Exp $
 *
 * libmnt.c : main() for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU Lesser General Public License 2.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include "libmnt.h"


static DBusConnection* connection = NULL;
static int is_initialized = FALSE;
static PLIBMNTFUNCS functions = NULL;
static void *user_data = NULL;

/**
 * @defgroup LibMnt MNT library
 * @brief A convenience library used to communicate with the MNT daemon
 *        using D-BUS
 *
 *  @{
 */


/** Message handler for method invocations. All invocations on any object
 *  or interface is routed through this function.
 *
 * @param  connection           D-BUS connection
 * @param  message              Message
 * @param  user_data            User data
 * @return                      What to do with the message
 * @internal
 */
static DBusHandlerResult filter_func(DBusConnection* connection,
                                     DBusMessage*    message,
                                     void*           user_data)
{
	DBusError error;
	
	g_assert(connection!=NULL);
	g_assert(message!=NULL);
	
	dbus_error_init(&error);

	if (dbus_message_is_signal(message, 
										DBUS_INTERFACE_MNT_MANAGER, 
										"VolumeMounted")) {
		char *udi = NULL;
		char *mntpnt = NULL;
		if (dbus_message_get_args(message, &error, 
											DBUS_TYPE_STRING, &udi,
											DBUS_TYPE_STRING, &mntpnt,
											DBUS_TYPE_INVALID)) {
			if (functions->volume_mounted!=NULL) {
				functions->volume_mounted(udi, mntpnt, user_data);
			}
			dbus_free(udi);
		}
		
	} else if (dbus_message_is_signal(message, 
										DBUS_INTERFACE_MNT_MANAGER, 
										"VolumeUnmounted")) {
		char *udi = NULL;
		char *mntpnt = NULL;
		if (dbus_message_get_args(message, &error, 
											DBUS_TYPE_STRING, &udi,
											DBUS_TYPE_STRING, &mntpnt,
											DBUS_TYPE_INVALID)) {
			if (functions->volume_unmounted!=NULL) {
				functions->volume_unmounted(udi, mntpnt, user_data);
			}
			dbus_free(udi);
		}
		
	}

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}


/** Dummy NULL callback functions */
static LIBMNTFUNCS dummy_functions = {
    NULL, /**< main_loop */
    NULL, /**< volume_mounted */
    NULL  /**< volume_unmounted */
};


/** Initialize the MNT library. 
 *
 * @param  cb_functions         Callback functions. If this is set to NULL
 *                              then the library will not listen for
 *                              notifications. 
 * @param  data                 User data pointer will be passed to callbacks
 * @return                      0 if successful, -1 otherwise
 */
int mnt_init(PLIBMNTFUNCS cb_functions, void *data)
{
	DBusError error;
	int res = 0;
	
	// exit, if already initialized
	if (is_initialized) {
		fprintf(stderr, "libmnt is already initialized!\n");
		goto cleanup;
	}

	// initialize the callback functions
	functions = cb_functions;
	if (cb_functions==NULL) {
		functions = &dummy_functions;
	}

	// initialize user_data
	user_data = data;
	
	// connect to dbus daemon and register myself
	dbus_error_init(&error);
	connection = dbus_bus_get(DBUS_BUS_SYSTEM, &error);
	if (connection==NULL) {
		fprintf(stderr, "Couldn't connect to DBUS, error %s: %s", 
			error.name, error.message);
		goto error;
	}
	
	// get my own service name
	//printf("Connected to DBUS_BUS_SYSTEM as '%s'.\n", dbus_bus_get_base_service(connection));
	
	// test for running service from mntd
	if (!dbus_bus_service_exists(connection, DBUS_SERVICE_MNT, &error)) {
		fprintf(stderr, "Service '%s' doesn't exist !\n", DBUS_SERVICE_MNT);
		goto error;
	} else {
		//printf("Service '%s' is running.\n", DBUS_SERVICE_MNT);
	}
	
	if (dbus_bus_acquire_service(connection, DBUS_SERVICE_MNT, 0, &error) == -1) {
		fprintf(stderr, "Error Acquiring service '" DBUS_SERVICE_MNT "'\n");
		if (dbus_error_is_set(&error)) {
			fprintf(stderr, "Can't acquire '" DBUS_SERVICE_MNT "' service, error %s: %s'",
				error.name, error.message);
		}
		goto error;
	} else {
		//printf("Aquiring service successful.\n");
	}
	
	// call main_loop function, if needed
	if (functions->main_loop!=NULL) {
		functions->main_loop(connection, user_data);
	}

	// add filter functions to connection handler ....
	if (!dbus_connection_add_filter(connection, filter_func, user_data, NULL)) {
		fprintf(stderr, "Couldn't add filters to dbus connection handler");
		goto error;
	} else {
		//printf("Filter added.\n");
	}
	// ..., but register only for signals from MNT daemon
	dbus_bus_add_match(connection, 
							"type='signal',"
							"interface='" DBUS_INTERFACE_MNT_MANAGER "',"
							"sender='" DBUS_SERVICE_MNT "',"
							"path='" DBUS_PATH_MNT_MANAGER "'",
							&error);
	if (dbus_error_is_set(&error)) {
		fprintf(stderr, "Couldn't register Manager signal handler: %s: %s", error.name, error.message);
		goto error;
	} else {
		//printf("Manager match added.\n");
	}

/*	dbus_bus_add_match(connection, 
							"type='signal',"
							"interface='" DBUS_INTERFACE_MNT_VOLUME "',"
							"sender='" DBUS_SERVICE_MNT "'", 
							&error);
	if (dbus_error_is_set(&error)) {
		fprintf(stderr, "Couldn't register Volume signal handler: %s: %s", error.name, error.message);
		goto error;
	} else {
		printf("Volume match added.\n");
	}
	*/
	//printf("initialized.\n");
	is_initialized = TRUE;
	
	goto cleanup;
	
error:
	dbus_connection_disconnect(connection);
	is_initialized = FALSE;
	res = -1;
	
cleanup:
	if (dbus_error_is_set(&error)) {
		dbus_error_free(&error);
	}

	return res;
}

/** Shutdown the MNT library.
 *
 * @return                      0 if successful, -1 otherwise
 */
int mnt_quit(void)
{
	if (!is_initialized) {
		return -1;
	}
	
	dbus_connection_disconnect(connection);
	connection = NULL;
	
	is_initialized = FALSE;
	return 0;
}


/** Get all volumes from mntd (with dbus method call). You have to free
 * each element and the data array by yourself.
 *
 * @param num_volumes           Integer pointer will hold number of found 
 *                              volumes, if successful
 * @return                      data pointer if successful, NULL otherwise
 */
char **mnt_get_all_volumes(int *num_volumes)
{
	DBusError error;
	DBusMessage* message = NULL;
	DBusMessage* reply = NULL;
	DBusMessageIter iter;
	char **res = NULL;
	char **data_array;
	int data_length;
	int i=0;
	
	message = dbus_message_new_method_call(
											DBUS_SERVICE_MNT,
											DBUS_PATH_MNT_MANAGER,
											DBUS_INTERFACE_MNT_MANAGER,
											"GetAllVolumes");
	
	if (message==NULL) {
		goto error;
	}
	
	dbus_error_init(&error);
	
	reply = dbus_connection_send_with_reply_and_block(connection, message, -1, &error);
	if (dbus_error_is_set(&error)) {
		goto error;
	}
	if (reply==NULL) {
		goto error;
	}
	
	// get reply data
	dbus_message_iter_init(reply, &iter);
	if (!dbus_message_iter_get_string_array(&iter, &data_array, &data_length)) {
		fprintf(stderr, "wrong reply from mntd\n");
		goto error;
	}
	
	// handle data
	res = (char **) malloc(sizeof(char*)*data_length);
	if (res == NULL) {
		goto error;
	}

	// fill with data
	*num_volumes = 0;
	for(i=0; i<data_length; i++) {
		if ((data_array[i] != NULL) && (strlen(data_array[i])>0)) {
			res[i] = strdup(data_array[i]);
			(*num_volumes)++;
		}
	}

	// free dbus data array
	dbus_free_string_array(data_array);
	
	goto cleanup;
	
error:
	res = NULL;
	
cleanup:
	if (message!=NULL) dbus_message_unref(message);
	if (reply!=NULL) dbus_message_unref(reply);
	
	return res;
}


/** Get mount point for a specific volume. You have to free to returned
 * mount point string yourself.
 *
 * @param udi                   udi of volume
 * @return                      mount point if successful, NULL otherwise
 */
char *mnt_get_mntpnt(char *udi)
{
	DBusError error;
	DBusMessage* message = NULL;
	DBusMessage* reply = NULL;
	DBusMessageIter iter;
	char *res = NULL;
	char *mntpnt = NULL;
	
	message = dbus_message_new_method_call(
											DBUS_SERVICE_MNT,
											DBUS_PATH_MNT_MANAGER,
											DBUS_INTERFACE_MNT_MANAGER,
											"GetMntPnt");
	
	if (message==NULL) {
		goto error;
	}
	
	// construct parameter
	dbus_message_iter_init(message, &iter);
	dbus_message_iter_append_string(&iter, udi);
	
	// method invocation
	dbus_error_init(&error);
	reply = dbus_connection_send_with_reply_and_block(connection, message, -1, &error);
	if (dbus_error_is_set(&error)) {
		fprintf(stderr, "Error calling GetMntPnt('%s'): %s : %s\n", udi, error.name, error.message);
		goto error;
	}
	if (reply==NULL) {
		goto error;
	}
	
	// get return value
	if (!dbus_message_get_args(reply, &error, 
										DBUS_TYPE_STRING, &mntpnt,
										DBUS_TYPE_INVALID)) {
		fprintf(stderr, "Error getting mntpnt from GetMntPnt() call (%s : %s)\n", error.name, error.message);
		goto error;
	}
	res = strdup(mntpnt);
	if (res == NULL) {
		goto error;
	}
	
	goto cleanup;
	
error:
	res = NULL;
	
cleanup:
	if (mntpnt!=NULL) dbus_free(mntpnt);
	if (message!=NULL) dbus_message_unref(message);
	if (reply!=NULL) dbus_message_unref(reply);
	
	return res;
}


/** @} */
