/**
 * Error Manager
 *
 * @defgroup error Error Manager
 * @author Stefan Bambach (stefan@bambach.biz)
 * @version 0.1
 * @{
 */

#ifndef __ERRMANAGER_H__
#define __ERRMANAGER_H__

#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

#include <syslog.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#define MAX_STRLEN 1024       /**< Maximum length of handled strings */

#define EM_TYPE_SYSLOG 1      /**< Log message to syslogd */
#define EM_TYPE_STDERR 2      /**< Log message to stderr */
#define EM_TYPE_CALLBACK 4    /**< Log message with callback function */
#define EM_TYPE_FILE 8        /**< Log message to file */
   
   /**
    * @def MSG_DEBUG        Output a DEBUG Message, when the "logLevel" is minimum "LOG_DEBUG"
    */
#ifdef WITH_DEBUG
	#define MSG_DEBUG(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_DEBUG, ## args)
#else
	#define MSG_DEBUG(args...) {}
#endif
   /**
    * @def MSG_INF         Output a INFO Message, when the "logLevel" is minimum "LOG_INFO"
    */
   #define MSG_INF(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_INFO, ## args)
   
   /**
    * @def MSG_NOTICE       Output a NOTICE Message, when the "logLevel" is minimum "LOG_NOTICE"
    */
   #define MSG_NOTICE(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_NOTICE, ## args)
   
   /**
    * @def MSG_WARNING      Output a WARNING Message, when the "logLevel" is minimum "LOG_WARNING"
    */
   #define MSG_WARNING(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_WARNING, ## args)
   
   /**
    * @def MSG_ERR          Output a ERROR Message, when the "logLevel" is minimum "LOG_ERR"
    */
   #define MSG_ERR(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_ERR, ## args)
   
   /**
    * @def MSG_CRIT         Output a CRITICAL Message, when the "logLevel" is minimum "LOG_CRIT"
    */
   #define MSG_CRIT(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_CRIT, ## args)
   
   /**
    * @def MSG_ALERT        Output a ALERT Message, when the "logLevel" is minimum "LOG_ALERT"
    */
   #define MSG_ALERT(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_ALERT, ## args)
      
   /**
    * @def MSG_EMERG        Output a EMERGENCY Message, when the "logLevel" is minimum "LOG_EMERG"
    */
   #define MSG_EMERG(args...) emManager(__FILE__, __FUNCTION__, __LINE__, LOG_EMERG, ## args)
      
   

   /**
    * Strukt for handling each msg
    */
   typedef struct EMDATA {
      int logLevel;         /**< Store loglevel */
      int logType;          /**< Store type of logging */
		
      /** Callback function that will be called, if logtype is EM_TYPE_CALLBACK
		 *
		 * @param ctxt        Context pointer
		 * @param pFile       Filename to log to, if logtype is EM_TYPE_FILE
		 * @param pFunc       Function name, where the message is generated
		 * @param iLine       Line number, where the message is generated
		 * @param level       Minimum loglevel that will be logged
		 * @param fmt         Format string of message
		 */
		void (* errCallBack)(void *ctxt, char *pFile, char *pFunc, int iLine, int level, char *fmt);
		
      void *ctxt;           /**< Context pointer */
      char *filename;       /**< Filename to log to, if logtype is EM_TYPE_FILE */
   } emData;
   
   
   /**
    * Init the Error Manager
    *
    * @param logLevel       use on of this LOG_DEBUG | LOG_INFO | LOG_NOTICE | 
    *                       LOG_WARNING | LOG_ERR | LOG_CRIT | LOG_ALERT | LOG_EMERG
    * @param logType        use EM_TYPE_SYSLOG for SYSLOG and EMTYPE_STDERR for stderr
    * @param cb             callback function for self defined output handling
    * @param ctxt           user defined data pointer (first parameter for callback function)
    * @param filename       filename to log to, if not using syslogd
    * @return               the logType or the logLevel or 0
    * @note                 When you do not init, syslog is used. Do not use 
    *                       this function, use the define INIT_STDERR
    * @todo                 loggin in a File specific by filename
    */
   int emInit(int logLevel, int logType, void (*cb)(void *ctxt, char *pFile, char *pFunc, int iLine, int level, char *fmt), void *ctxt, char *filename);
   
   
   /**
    * print messages to selected log medium
    * 
    * @param pFile          Filename of the function who has called the ErrorManager
    * @param pFunc          Function name who has called the ErrorManager
    * @param iLine          Linenumber of the function who has called the ErrorManager
    * @param level          The Loglevel
    * @param fmt            The output String
    * @return               void
    *
    * @note:               Do not use this function, use the defines.
    *
    * @internal
    */
	void emManager(char* pFile, char* pFunc, int iLine, int level, char* fmt, ...);
   
   
#endif

/** @} */
