#include "hashmap.h"
#include "dbug_mem.h"


int
chtbl_init(CHTbl *htbl, int buckets, int (*h)(const void *key), int
   (*match)(const void *obj1, const void *obj2), void (*destroy)(void *obj))
{
	int i;

	//MSG_DEBUG("chtbl_init() called");
	
	T_RDWR_INIT(&(htbl->rwlock), NULL);

	/* Allocate space for the hash table. */
	if ((htbl->table = (List *)malloc(buckets * sizeof(List))) == NULL)
		return -1;

	/* Initialize the buckets. */
	htbl->buckets = buckets;

	for (i = 0; i < htbl->buckets; i++) {
		list_init(&htbl->table[i], match, destroy);
	}

	/* Encapsulate the functions. */
	htbl->h = h;

	return 0;
}


void
chtbl_destroy(CHTbl *htbl)
{
	int i;

	//MSG_DEBUG("chtbl_destroy() called");
	
	T_RDWR_WLOCK(&(htbl->rwlock));

	/* Destroy each bucket. */
	for (i = 0; i < htbl->buckets; i++) {
		list_destroy(&htbl->table[i]);
	}
	
	/* Free the storage allocated for the hash table. */
	free(htbl->table);
	
	
	/* No operations are allowed now, but clear the structure as a precaution. */
	memset(htbl, 0, sizeof(CHTbl));
	
	T_RDWR_WUNLOCK(&(htbl->rwlock));

	return;
}


int
chtbl_size(CHTbl *htbl)
{
	int cnt = 0;
	int i;
	
	//MSG_DEBUG("chtbl_size() called");
	
	T_RDWR_RLOCK(&(htbl->rwlock));
	
	for (i = 0; i < htbl->buckets; i++) {
		cnt += list_getsize(&htbl->table[i]);
	}
	
	T_RDWR_RUNLOCK(&(htbl->rwlock));
	
	return cnt;
}


int
chtbl_insert(CHTbl *htbl, const void *data)
{
	void *temp;
	int bucket, retval;
	 

	//MSG_DEBUG("chtbl_insert() called");
	
	/* Do nothing if the data is already in the table. */
	temp = (void *)data;
	
	T_RDWR_RLOCK(&(htbl->rwlock));

	/* Hash the key. */
	bucket = htbl->h(data) % htbl->buckets;
	
	/* Insert the data into the bucket. */
	retval = list_insert(&htbl->table[bucket], (void *)data);
	
	T_RDWR_RUNLOCK(&(htbl->rwlock));

	return retval;
}


int
chtbl_remove(CHTbl *htbl, void **data)
{
	int bucket;
	int retval = -1;
 
	//MSG_DEBUG("chtbl_remove() called");
	
	T_RDWR_RLOCK(&(htbl->rwlock));

	/* Hash the key. */
	bucket = htbl->h(*data) % htbl->buckets;
	
	/* Remove the data from the bucket. */
	retval = list_remove(&htbl->table[bucket], data);

	T_RDWR_RUNLOCK(&(htbl->rwlock));

	/* Return that the data was not found. */
	return retval;
}


int
chtbl_lookup(CHTbl *htbl, void **data)
{
	int bucket = 0;
	int retval = 0;
	
	//MSG_DEBUG("chtbl_lookup() called");
	
	T_RDWR_RLOCK(&(htbl->rwlock));

	/* Hash the key. */
	bucket = htbl->h(*data) % htbl->buckets;
	
	retval = list_lookup(&(htbl->table[bucket]), data);
	
	T_RDWR_RUNLOCK(&(htbl->rwlock));

	return retval;
}
