/**
 * Hash map handling
 *
 * @defgroup chash Chained Hash Map
 * @author Stefan Bambach (sbambach@gmx.net)
 * @version 0.1
 * @{
 */

#ifndef __CHASH_CHTBL_H__
#define __CHASH_CHTBL_H__

#ifdef HAVE_CONFIG_H
	#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "threading.h"
#include "errmanager.h"
#include "linklist.h"


/*
 * public defines
 */




/*
 * public structs
 */


/**
 * Structure for the hashmap.
 * 
 * Holds the internal data for managing the hashmap.
 */
typedef struct CHTbl_ {
	/** Struct for Locking the table (read/write lock) */
	T_RDWR rwlock;
	
	/** Number of buckets to use */
	int buckets;
	
	/** Hash function */
	int (*h)(const void *key);
	
	/** Pointer to hash map entry */
	List *table;
	
} CHTbl;


/*
 * public methods
 */


/**
 * Initialize hash map
 *
 * @param htbl			Pointer to Hash table struct
 * @param buckets		No. of Buckets to use
 * @param h				hash function (calculate Bucketno. fro CacheObj)
 * @param match		match function (compare two CacheObjs)
 * @param destroy		destroy function (destroy CacheObj)
 * @return				0 if successful, -1 otherwise.
 */
int chtbl_init(CHTbl *htbl, int buckets, int (*h)(const void *key), int
   (*match)(const void *obj1, const void *obj2), void (*destroy)(void *obj));


/**
 * Destroy hash map
 *
 * @param htbl			Pointer to Hash table struct
 */
void chtbl_destroy(CHTbl *htbl);


/**
 * Destroy hash map
 *
 * @param htbl			Pointer to Hash table struct
 * @return				Number of elements in hash map
 */
int chtbl_size(CHTbl *htbl);


/**
 * Insert element in hash map
 *
 * @param htbl			Pointer to Hash table struct
 * @param data			Pointer to object to be stored
 * @return				0 if successfully inserted
 *						-1 if error while inserting
 *						1 if element is already in list
 */
int chtbl_insert(CHTbl *htbl, const void *data);


/**
 * Remove element from hash map
 *
 * @param htbl			Pointer to Hash table struct
 * @param data			Pointer to pointer to object to be removed
 * @return				0 if successfully removed, -1 otherwise.
 */
int chtbl_remove(CHTbl *htbl, void **data);


/**
 * Lookup for element in hash map
 *
 * @param htbl			Pointer to Hash table struct
 * @param data			Pointer to pointer to object to be searched for
 * @return				0 if element found, -1 otherwise.
 */
int chtbl_lookup(CHTbl *htbl, void **data);


/*
 * private methods
 */




#endif

/** @} */
