/**
 * Link list handling for hash map
 *
 * @defgroup list Single Linked List
 * @ingroup chash
 * @author Stefan Bambach (sbambach@gmx.net)
 * @version 0.1
 * @{
 */

#ifndef __CHASH_LIST_H__
#define __CHASH_LIST_H__

#ifdef HAVE_CONFIG_H
	#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "threading.h"
#include "errmanager.h"

/*
 * public defines
 */


#define list_size(list) ((list)->size)       /**< get list size */
#define list_head(list) ((list)->head)       /**< get first element */
#define list_tail(list) ((list)->tail)       /**< get last element */
#define list_is_head(list, element) ((element) == (list)->head ? 1 : 0) /**< is it head of list */
#define list_is_tail(element) ((element)->next == NULL ? 1 : 0)         /**< is it tail of list */
#define list_data(element) ((element)->data) /**< get data of element */
#define list_next(element) ((element)->next) /**< get next element */


/*
 * public structs
 */


/** Define a structure for linked list elements.
 *
 */
typedef struct ListElmt_ {
	void				*data;      /**< elements data pointer */
	struct ListElmt_	*next;   /**< next element pointer */
} ListElmt;


/** Define a structure for linked lists.
 *
 */
typedef struct List_ {
	T_RDWR      rwlock;			/**< Struct for Locking */
	int         size;			   /**< No. of elements in this list */

	/** Callback Match function
	 *
	 * Compare two objects and return the result
	 * 
	 * @param key1             Object 1
	 * @param key2             Object 2
	 * @returns                1 if objects are equal, and 0 if not
	 */
	int         (*match)(const void *key1, const void *key2);

	/** Destroy function for one element
	 *
	 * @param data             destroy element identified by data pointer
	 */
	void        (*destroy)(void *data);

	ListElmt    *head;			/**< Pointer to first element */
	ListElmt    *tail;			/**< Pointer to last element */
} List;


/*
 * public methods
 */


/**
 * Initialize link list
 *
 * @param list       Pointer to List struct
 * @param match      Match function
 * @param destroy    Destroy function
 * @return           0 if successfully removed, -1 otherwise.
 */
void list_init(List *list, int (*match)(const void *key1, const void *key2), void (*destroy)(void *data));


/**
 * Destroy link list
 *
 * @param list       Pointer to List struct
 * @return           0 if successfully removed, -1 otherwise.
 */
void list_destroy(List *list);


/**
 * Get number of elements in list
 *
 * @param list       Pointer to List struct
 * @return           Number of elements in link list
 */
int list_getsize(List *list);


/**
 * Insert element in link list
 *
 * @param list       Pointer to List struct
 * @param data       Pointer to data to be removed
 * @return           0 if successfully inserted, -1 otherwise.
 */
int list_insert(List *list, void *data);


/**
 * Lookup for element in link list
 *
 * @param list       Pointer to List struct
 * @param data       Pointer to pointer to data to be searched
 * @return           0 if successfully removed, -1 otherwise.
 */
int list_lookup(List *list, void **data);

/**
 * Lookup for element in link list and call function for each match
 *
 * @param list       Pointer to List struct
 * @param matchfunc  Pointer to callback-function that will decide, 
 *                   if cbfunc will be called
 * @param cbfunc     Pointer to callback-function
 * @param data       Pointer to pointer to data to be searched
 * @return           0 if successfully removed, -1 otherwise.
 */
int list_foreach_match(List *list, int (*matchfunc)(const void *key1, const void *key2), void (*cbfunc)(const void *data), void **data);


/**
 * Iterate over each element of link list and call function
 *
 * @param list       Pointer to List struct
 * @param func       This function is called for each element in list
 * @param userdata   This is an argument for the callback function
 * @return           0 if successfully called function for all elements,
 *                   -1 otherwise.
 */
int list_for_each_call(List *list, int (*func)(void *data, void *userdata), void *userdata);


/**
 * Remove each element from link list, where function returns 1
 *
 * @param list       Pointer to List struct
 * @param func       This function is called for each element in list
 * @param userdata   This is an argument for the callback function
 * @return           0 if successfully called function for all elements,
 *                   -1 otherwise.
 */
int list_remove_each_if_func_eq_1(List *list, int (*func)(void *data, void *userdata), void *userdata);


/**
 * Remove element from link list
 *
 * @param list       Pointer to List struct
 * @param data       Pointer to pointer to data to be removed
 * @return           0 if successfully removed, -1 otherwise.
 */
int list_remove(List *list, void **data);


/*
 * private methods
 */


/**
 * Insert element in link list
 *
 * @param list       Pointer to List struct
 * @param element    Pointer to list element
 * @param data       Pointer to data to be stored
 * @return           0 if successfully removed, -1 otherwise.
 * @internal
 */
int list_ins_next(List *list, ListElmt *element, const void *data);


/**
 * Remove element from link list
 *
 * @param list       Pointer to List struct
 * @param element    Pointer to list element
 * @param data       Pointer to pointer to data to be removed
 * @return           0 if successfully removed, -1 otherwise.
 * @internal
 */
int list_rem_next(List *list, ListElmt *element, void **data);


/**
 * Simple match function for strings
 *
 * @param key1       First String
 * @param key2       Second String
 * @return           1 if equal, 0 otherwise.
 * @internal
 */
int stringMatchFunction(const void *key1, const void *key2);


/**
 * Simple free function for strings
 *
 * @param ptr        Pointer to char buffer to be freed
 * @internal
 */
void stringFreeFunction(void *ptr);


/**
 * Lookup for element in link list
 *
 * @param list       Pointer to List struct
 * @param data       Pointer to pointer to data to be searched
 * @return           0 if successfully removed, -1 otherwise.
 * @internal
 */
int list_internal_lookup(List *list, void **data);


#endif


/** @} */
