/***************************************************************************
 * CVSID: $Id: main.c,v 1.20 2004/05/25 21:20:24 stefanb Exp $
 *
 * main.c : main() for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libmnt/libmnt.h>   /* for common defines etc. */

#define _GNU_SOURCE
#include <getopt.h>

#include "errmanager.h"
#include "mntd_hal.h"
#include "mntd_file.h"
#include "mntd_dbus_manager.h"
#include "mntd_volume_manager.h"
#include "main.h"

/* global pointer to volume manager */
PVOLUMEMANAGER vols = NULL;

void usage(void);


/**
 * @defgroup MntDaemon MNT daemon
 * @brief The MNT daemon manages persistent volume objects available through
 *        a D-BUS network API
 * @{
 */
                                                                                                                           

/** Show usage
 */
void 
usage(void)
{
	printf("Mount Daemon " PACKAGE_VERSION "\n");
	printf("Mounts devices automatically and unmount it again. It's meant\n");
	printf("to handle hotplug devices like USB storage or similar stuff.\n");
	printf("\n");
	printf("usage: mntd [-h] [-d] [-f configfile]\n");
	printf("\n");
	printf("    -h              Help (this text)\n");
	printf("    -d              Daemonize it\n");
	printf("    -f,--file       Use given configfile\n");
	printf("\n");
}


/** Entry point for HAL daemon
 *
 *  @param  argc                Number of arguments
 *  @param  argv                Array of arguments
 *  @return                     Exit code
 */
int main(int argc, char* argv[])
{
	GMainLoop* loop;
	char *config = NULL;
	int i=0;
	char *configfile = NULL;
	char *configfiles[] = {
		MNTD_CONF_DIR,
		"/etc/mntd/mntd.conf",
		"/etc/mntd.conf",
		"./mntd.conf",
		"../mntd.conf",
		NULL
	};
	int daemonize = FALSE;
	
	// parse commandline options
	while(1) {
		int c;
		int option_index = 0;
		static struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"daemonize", no_argument, 0, 'd'},
			{"file", required_argument, 0, 'f'},
			{0, 0, 0, 0}
		};
		c = getopt_long (argc, argv, "hf:d", long_options, &option_index);
		if (c == -1) {
			break;
		}
		switch (c) {
		case 'd':
			daemonize = TRUE;
			break;
		
		case 'f':
			config = optarg;
			break;
		
		case 'h':
			usage();
			exit(0);
			break;
		
		default:
			usage();
			exit(1);
			break;
		}
	}
	
	// initialize error manager
	emInit(LOG_DEBUG, EM_TYPE_SYSLOG, NULL, NULL, NULL);
	
	// show version
	MSG_INF("MNT daemon version " PACKAGE_VERSION " starting up");
	
	// try to load config file
	if (config == NULL) {
		MSG_DEBUG("No config file specified. trying to find it.");
		i=0;
		while (configfiles[i]!=NULL) {
			configfile = configfiles[i];
			MSG_DEBUG("trying '%s' ...", configfile);
			if (mntd_file_is_file(configfile)) {
				MSG_DEBUG("found '%s'", configfile);
				config = configfile;
				break;
			}
			i++;
		}
	}
	
	if (config == NULL) {
		MSG_EMERG("No config file given and nothing found. exiting.");
		exit(1);
	}
	if (!mntd_file_is_file(config)) {
		MSG_EMERG("Config file '%s' cannot be accessed.", config);
		exit(1);
	}
	
	// handle daemon mode
	if (daemonize==TRUE) {
		MSG_INF("daemonize not implemented yet -> ignored");
	}
	
	// start volume manager
	vols = new_VolumeManager(config);
	if (vols == NULL) {
		MSG_EMERG("Couldn't not initialize system.");
		exit(1);
	}
	
	loop = g_main_loop_new(NULL, FALSE);
	
	// initialize my dbus stuff
	mntd_dbus_init(vols);
	
	// start HAL handler
	mntd_hal_init();
	
	// Scan for Volumes
	vols->rescan(vols);
	
	// run the main loop
	g_main_loop_run(loop);
	
	// destroy HAL handler
	mntd_hal_shutdown();
	
	// destroy dbus handler
	mntd_dbus_quit();
	
	// free volume storage
	g_assert(vols!=NULL);
	vols->destroy(vols);
	
	// MNT daemon exists
	MSG_INF("MNT daemon version " PACKAGE_VERSION " exited.");

	return 0;
}

/** @} */
