/***************************************************************************
 * CVSID: $Id: mntd_dbus_manager.c,v 1.19 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_dbus_manager.c : D-BUS manager for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include <libmnt/libmnt.h>   /* for common defines etc. */

#include "mntd_dbus_error.h"
#include "mntd_dbus_manager.h"
#include "mntd_volume_manager.h"
#include "mntd_volume.h"
#include "errmanager.h"
#include "main.h"


static void *user_data = NULL;


// function prototypes
int cb_append_udi(void *data, void *userdata);


/***************************************************************************
 * MANAGER
 **************************************************************************/

/**
 * @defgroup DaemonManager D-BUS Manager
 * @ingroup MntDaemon
 * @brief Manager for D-BUS messages for MNT Daemon
 * @{
 */


/** D-Bus connection object for the MNT service */
DBusConnection* dbus_connection = NULL;
                                                                                                                           

/** Initialize DBUS stuff for MNT daemon (biz.bambach.Mnt)
 *
 * @return                      0 if successful, -1 otherwise
 */
int 
mntd_dbus_init(void *data)
{
	DBusError dbus_error;

	dbus_connection_set_change_sigpipe(TRUE);
	
	dbus_error_init(&dbus_error);
	dbus_connection = dbus_bus_get(DBUS_BUS_SYSTEM, &dbus_error);
	if (dbus_connection==NULL) {
		MSG_ERR("dbus_bus_get(): '%s'", dbus_error.message);
		return -1;
	}
	
	dbus_connection_setup_with_g_main(dbus_connection, NULL);
	
	dbus_bus_acquire_service(dbus_connection, DBUS_SERVICE_MNT, 0, &dbus_error);
	if (dbus_error_is_set(&dbus_error)) {
		MSG_ERR("dbus_bus_acquire_service(): '%s'", dbus_error.message);
		return -1;
	}
	
	user_data = data;
	
	dbus_connection_add_filter(dbus_connection, mntd_dbus_manager_filter_function, 
										user_data, NULL);

	return 0;
}


/** Quit DBUS for MNT daemon
 *
 * @return                      0 if successful, -1 otherwise
 */
int 
mntd_dbus_quit(void)
{
	return 0;
}


/** Raise the biz.bambach.Mnt.NoSuchDevice error
 *
 *  @param  connection          D-Bus connection
 *  @param  in_reply_to         message to report error on
 *  @param  udi                 Unique device id given
 */
void mntd_dbus_raise_no_such_device(DBusConnection* connection,
											DBusMessage* in_reply_to,
											const char* udi)
{
	char buf[512];
	DBusMessage *reply;
	
	snprintf(buf, 511, "No device with id %s", udi);
	//MSG_ERR(buf);
	
	reply = dbus_message_new_error(in_reply_to, 
											DBUS_ERROR_NO_SUCH_DEVICE, 
											buf);
	if (reply==NULL) {
		MSG_EMERG("Out of memory");
		return;
	}
	if (!dbus_connection_send(connection, reply, NULL)) {
		MSG_EMERG("Out of memory");
		return;
	}
	dbus_message_unref(reply);
}


/** Raise the biz.bambach.Mnt.SyntaxError error
 *
 *  @param  connection          D-Bus connection
 *  @param  in_reply_to         message to report error on
 *  @param  method_name         Name of the method that was invoked with
 *                              the wrong signature
 */
void mntd_dbus_raise_syntax(DBusConnection* connection,
											DBusMessage* in_reply_to,
											const char* method_name)
{
	char buf[512];
	DBusMessage *reply;
	
	snprintf(buf, 511, 
				"There is a syntax error in the invocation of "
				"the method %s", method_name);
	//MSG_ERR(buf);
	
	reply = dbus_message_new_error(in_reply_to, 
											DBUS_ERROR_SYNTAX, 
											buf);
	if (reply==NULL) {
		MSG_EMERG("Out of memory");
		return;
	}
	if (!dbus_connection_send(connection, reply, NULL)) {
		MSG_EMERG("Out of memory");
		return;
	}
	dbus_message_unref(reply);
}


/** Raise the biz.bambach.Mnt.NoMntPnt error
 *
 *  @param  connection          D-Bus connection
 *  @param  in_reply_to         message to report error on
 *  @param  udi                 Unique device id given
 */
void mntd_dbus_raise_no_mntpnt(DBusConnection* connection,
											DBusMessage* in_reply_to,
											const char* udi)
{
	char buf[512];
	DBusMessage *reply;
	
	snprintf(buf, 511, 
				"No mount point for '%s' found.", udi);
	//MSG_ERR(buf);
	
	reply = dbus_message_new_error(in_reply_to, 
											DBUS_ERROR_NO_MNTPNT, 
											buf);
	if (reply==NULL) {
		MSG_EMERG("Out of memory");
		return;
	}
	if (!dbus_connection_send(connection, reply, NULL)) {
		MSG_EMERG("Out of memory");
		return;
	}
	dbus_message_unref(reply);
}


/** Message handler for method invocations. All invocations on any object
 *  or interface is routed through this function. All method calls for MNT 
 *  daemon will be handled here.
 *
 *  @param  connection          D-BUS connection
 *  @param  message             Message
 *  @param  user_data           User data
 *  @return                     What to do with the message
 */
DBusHandlerResult mntd_dbus_manager_filter_function(DBusConnection* connection,
                                         DBusMessage* message,
                                         void* user_data)
{
	if (dbus_message_is_method_call(message,
											DBUS_INTERFACE_MNT_MANAGER,
											"GetAllVolumes")) {
		MSG_DEBUG("biz.bambach.Mnt.Manager::GetAllVolumes() called");
		return mntd_dbus_manager_get_all_volumes(vols, connection, message);
												
	} else if (dbus_message_is_method_call(message,
											DBUS_INTERFACE_MNT_MANAGER,
											"GetMntPnt")) {
		MSG_DEBUG("biz.bambach.Mnt.Manager::GetMntPnt() called");
		return mntd_dbus_manager_get_mntpnt(vols, connection, message);
												
/*	} else {
		int type;
		type = dbus_message_get_type(message);
		switch(type) {
		case DBUS_MESSAGE_TYPE_METHOD_CALL:
			MSG_DEBUG("DBUS_MESSAGE_TYPE_METHOD_CALL");
			break;
		case DBUS_MESSAGE_TYPE_METHOD_RETURN:
			MSG_DEBUG("DBUS_MESSAGE_TYPE_METHOD_RETURN");
			break;
		case DBUS_MESSAGE_TYPE_ERROR:
			MSG_DEBUG("DBUS_MESSAGE_TYPE_ERROR");
			break;
		case DBUS_MESSAGE_TYPE_SIGNAL:
			MSG_DEBUG("DBUS_MESSAGE_TYPE_SIGNAL");
			break;
		case DBUS_MESSAGE_TYPE_INVALID:
			MSG_DEBUG("DBUS_MESSAGE_TYPE_INVALID");
			break;
		default:
			MSG_DEBUG("DBUS_MESSAGE_TYPE_ .... UNKNOWN");
			break;
		}
		
		MSG_INF("obj_path=%s interface=%s method=%s destination=%s sender=%s", 
					dbus_message_get_path(message), 
					dbus_message_get_interface(message),
					dbus_message_get_member(message),
					dbus_message_get_destination(message),
					dbus_message_get_sender(message)
		);*/
	}
	
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}


/** Send signal VolumeAdded(string udi) on the biz.bambach.Mnt.Manager
 *  interface on the object /biz/bambach/Mnt/Manager.
 *
 * @param pv                    Pointer to Volume Object
 */
void 
mntd_dbus_manager_send_signal_volume_mounted(PVOLUME pv)
{
	DBusMessage* message;
	DBusMessageIter it;
	
	//MSG_DEBUG("mntd_dbus_manager_send_signal_volume_mounted() called");

	message = dbus_message_new_signal(
											DBUS_PATH_MNT_MANAGER, 
											DBUS_INTERFACE_MNT_MANAGER,
											"VolumeMounted");
	if (message==NULL) {
		MSG_EMERG("Out of memory");
		return;
	}
	
	dbus_message_iter_init(message, &it);
	dbus_message_iter_append_string(&it, pv->udi);
	dbus_message_iter_append_string(&it, pv->mntpnt);
	
	if (!dbus_connection_send(dbus_connection, message, NULL)) {
		MSG_EMERG("error broadcasting message");
	}
	
	dbus_message_unref(message);
}


/** Send signal VolumeRemoved(string udi) on the biz.bambach.Mnt.Manager
 *  interface on the object /biz/bambach/Mnt/Manager.
 *
 * @param pv                    Pointer to Volume Object
 */
void 
mntd_dbus_manager_send_signal_volume_unmounted(PVOLUME pv)
{
	DBusMessage* message;
	DBusMessageIter it;
	
	//MSG_DEBUG("mntd_dbus_manager_send_signal_volume_unmounted() called");

	message = dbus_message_new_signal(
											DBUS_PATH_MNT_MANAGER, 
											DBUS_INTERFACE_MNT_MANAGER,
											"VolumeUnmounted");
	if (message==NULL) {
		MSG_EMERG("Out of memory");
		return;
	}
	
	dbus_message_iter_init(message, &it);
	dbus_message_iter_append_string(&it, pv->udi);
	dbus_message_iter_append_string(&it, pv->mntpnt);
	
	if (!dbus_connection_send(dbus_connection, message, NULL)) {
		MSG_EMERG("error broadcasting message");
	}
	
	dbus_message_unref(message);
}


/** Callback for each entry in volumes
 *
 *  @param  data                PHASHELEMENT pointer
 *  @param  userdata            dbus iterator pointer
 *  @return                     0 if successful, -1 otherwise
 */
int cb_append_udi(void *data, void *userdata)
{
	DBusMessageIter *itarr = NULL;
	PHASHMAPELEMENT phme = NULL;
	PHASHMAP phm = NULL;
	PVOLUME pv = NULL;
	char *udi = NULL;
	void *hashdata = NULL;

	if ((data==NULL) || (userdata==NULL)) {
		return -1;
	}
	
	// get dbus iterator
	itarr = (DBusMessageIter *) userdata;
	// get volume hashmap element
	phme = (PHASHMAPELEMENT) data;
	// get hashmap object
	phm = phme->phm;
	// get hashmap elements key
	udi = phme->key;
	// get volume object
	if (phm->get(phm, udi, &hashdata) == 0) {
		pv = (PVOLUME) hashdata;
		if (pv != NULL) {
			// check for mntd handled mntpnt
			if (pv->is_in_mntpath(pv)) {
				// append it to dbus message iterator
				dbus_message_iter_append_string(itarr, udi);
				return 0;
			}
		}
	}

	return -1;
}


/** Get all volumes.
 *
 *  <pre>
 *  array{object_reference} Manager.GetAllVolumes()
 *  </pre>
 *
 *  @param pvm                  pointer to volume manager
 *  @param connection           D-BUS connection
 *  @param message              Message
 *  @return                     What to do with the message
 */
DBusHandlerResult 
mntd_dbus_manager_get_all_volumes(	PVOLUMEMANAGER pvm,
												DBusConnection* connection,
												DBusMessage* message)
{
	DBusMessage* reply;
	DBusMessageIter it;
	DBusMessageIter itarr;
	DBusHandlerResult res = DBUS_HANDLER_RESULT_HANDLED;
	
	// return data to library
	reply = dbus_message_new_method_return(message);
	if (reply==NULL) {
		MSG_EMERG("Out of memory");
		goto error;
	}
	
	// collect return data
	dbus_message_iter_init(reply, &it);
	dbus_message_iter_append_array(&it, &itarr, DBUS_TYPE_STRING);
	pvm->foreach_mounted(pvm, cb_append_udi, &itarr);
	
	// send it
	if(!dbus_connection_send(connection, reply, NULL)) {
		MSG_EMERG("Out of memory");
		goto error;
	}
	
	goto cleanup;
	
error:
	res = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	
cleanup:
	if (reply!=NULL) {
		dbus_message_unref(reply);
	}
	
	return res;
}


/** Get mount point.
 *
 *  <pre>
 *  string{object_reference} Manager.GetMntPnt()
 *  </pre>
 *
 *  @param pvm                  pointer to volume manager
 *  @param connection           D-BUS connection
 *  @param message              Message
 *  @return                     What to do with the message
 */
DBusHandlerResult 
mntd_dbus_manager_get_mntpnt(			PVOLUMEMANAGER pvm,
												DBusConnection* connection,
												DBusMessage* message)
{
	DBusMessage *reply = NULL;
	DBusMessageIter it;
	DBusError error;
	DBusHandlerResult res = DBUS_HANDLER_RESULT_HANDLED;
	char *udi = NULL;
	char *mntpnt = NULL;
	
	// get parameter
	dbus_error_init(&error);
	if (!dbus_message_get_args(message, &error, 
										DBUS_TYPE_STRING, &udi,
										DBUS_TYPE_INVALID) )
	{
		mntd_dbus_raise_syntax(connection, message, "Manager.GetMntPnt");
		goto error;
	}
	
	// return data to library
	reply = dbus_message_new_method_return(message);
	if (reply==NULL) {
		MSG_EMERG("Out of memory");
		goto error;
	}
	
	// collect return data
	dbus_message_iter_init(reply, &it);
	if (pvm->contains(pvm, udi)) {
		mntpnt = pvm->get_mntpnt(pvm, udi);
		if (mntpnt==NULL) {
			mntd_dbus_raise_no_mntpnt(connection, message, udi);
			goto error;
		}
		dbus_message_iter_append_string(&it, mntpnt);
	} else {
		mntd_dbus_raise_no_such_device(connection, message, udi);
		goto error;
	}
	
	// send it
	if(!dbus_connection_send(connection, reply, NULL)) {
		MSG_EMERG("Out of memory");
		goto error;
	}
	
	goto cleanup;
	
error:
	res = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	
cleanup:
	if (reply!=NULL) {
		dbus_message_unref(reply);
	}
	if (udi!=NULL) {
		dbus_free(udi);
	}
	if (mntpnt!=NULL) {
		free(mntpnt);
	}
	
	return res;
}


/** @} */


/***************************************************************************
 * VOLUME HANDLER
 **************************************************************************/

/**
 * @defgroup VolumeInterface D-BUS interface biz.bambach.Mnt.Volume
 * @ingroup MntDaemon
 * @brief D-BUS interface for generic volume operations
 * @{
 */


/** @} */
