/***************************************************************************
 * CVSID: $Id: mntd_dir.c,v 1.2 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_dir.c : directory handler for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>

#include "mntd_dir.h"
#include "mntd_file.h"
#include "errmanager.h"


extern int errno;


/**
 * @defgroup DaemonHelperDir Helper for directory handling
 * @ingroup MntDaemon
 * @brief Various helper functions for directories
 * @{
 */


/**
 * Check if path is a directory
 *
 * @param dirpath               Path to directory
 * @return                      1 if successful, and 0 otherwise
 */
int
mntd_dir_is_dir(char *dirpath)
{
	struct stat st;
	
	if (stat(dirpath, &st) != 0) {
		return 0;
	}

	if (!S_ISDIR(st.st_mode)) {
		return 0;
	}
	
	return 1;
}


/**
 * Remove one directory if it is empty 
 *
 * @param dirpath               Path to directory
 * @return                      1 if successful, and 0 otherwise
 */
int
mntd_dir_remove(char *dirpath)
{
	if(mntd_dir_is_dir(dirpath)) {
		if(rmdir(dirpath)==0) {
			return 1;
		}
	}
	
	return 0;
}


/**
 * Make directories and all missing parent directories
 *
 * @param pathname              Path to make
 * @param mode                  creating mask of permissions
 * @param error                 Pointer to error return value
 * @return                      0 if successful, and 1 otherwise
 */
int
mntd_dir_mkdirs(const char *pathname, mode_t mode, int *error)
{
	char path[FILENAME_MAX];
	char subpath[FILENAME_MAX];
	int i;
	int pathLength;
	int subpathLength = 0; 

	// check params
	if(pathname == NULL) {
		return -2;
	}

	// get length of pathname
	pathLength = strlen(pathname);

	// check about enough space for creating dirs
	if(pathLength > FILENAME_MAX - 2) {
		return -2;
	}

	// fill strings with \0
	memset(path, 0, FILENAME_MAX);
	memset(subpath, 0, FILENAME_MAX);

	// make working copy
	strncpy(path, pathname, FILENAME_MAX);

	// if DIRSEP not at the end of path, add it
	if(path[pathLength - 1] != DIRSEP) {
		path[pathLength] = DIRSEP;
	}

	// If dir exist thenwe have nothing to do
	if(mntd_dir_is_dir(path)) {
		return 0;
	}

	// If first char an DIRSEP then we need an other counter start
	if(path[0] != '/') {
		i=0;
	} else {
		subpath[0] = path[0];
		i=1;
	}

	// Go char for char thru the pathname an in case of a DIRSEP we check for DIR
	for(; path[i] != '\0'; i++) {
		subpath[i] = path[i];
		if(path[i] == DIRSEP) {
			subpathLength++;
			if(!mntd_dir_is_dir(subpath)) {
				if(mkdir(subpath, mode) == -1) {
					*error = errno;
					return -1;
				}
			}
		}
	}
	
	return 0;
}



/**
 * Remove directories and file recursivly
 *
 * @param dirname               Root directory from which file and 
 *                              directories will be removed
 *                              recursivly
 * @return                      1 if successful, and 0 otherwise
 */
int
mntd_dir_rmdirs(char *dirname)
{
	DIR *dir = NULL;
	struct dirent *dent = NULL;
	char *entry = NULL;
	char *rootdir = NULL;
	
	rootdir = strdup(dirname);
	if (rootdir == NULL) {
		return 0;
	}
		
	if (LAST_CHAR(rootdir)==DIRSEP) {
		LAST_CHAR(rootdir)='\0';
	}
		
	dir = opendir(rootdir);
	if (dir == NULL) {
		return 0;
	}
	
	while ((dent = readdir(dir)) != NULL) {
		if ((strcmp(dent->d_name, ".") != 0) && (strcmp(dent->d_name, "..") != 0)) {
			entry = (char *) malloc(sizeof(char)*(strlen(rootdir)+1+strlen(dent->d_name)+1));
			if (entry == NULL) {
				if (dir != NULL) {
					closedir(dir);
					dir = NULL;
				}
				return 0;
			}
			
			sprintf(entry, "%s/%s", rootdir, dent->d_name);
			
			if(mntd_file_is_file(entry)) {
				if(!mntd_file_remove(entry)) {
					;
				}
			} else if(mntd_dir_is_dir(entry)) {
				mntd_dir_rmdirs(entry);
				if(!rmdir(entry)) {
					;
				}
			} else {
				;
			}
			
			if (entry != NULL) {
				free(entry);
				entry = NULL;
			}
		}
	}
	
	if (dir != NULL) {
		closedir(dir);
		dir = NULL;
	}
	if (rootdir != NULL) {
		free(rootdir);
		rootdir = NULL;
	}

	return 1;
}


/** @} */
