/***************************************************************************
 * CVSID: $Id: mntd_file.c,v 1.1 2004/04/01 19:30:23 stefanb Exp $
 *
 * mntd_file.c : file handler for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include "mntd_file.h"
#include "errmanager.h"


/**
 * @defgroup DaemonHelperFile Helper for file handling
 * @ingroup MntDaemon
 * @brief Various helper functions for files
 * @{
 */


/**
 * Test if path is a FIFO
 *
 * @param filepath              Filepath
 * @return                      CTRUE if file is there, CFALSE otherwise
 * @internal
 */
int
mntd_file_is_fifo(char *filepath)
{
	struct stat st;
	
	if (stat(filepath, &st) != 0) {
		return 0;
	}

	if (!S_ISFIFO(st.st_mode)) {
		return 0;
	}
	
	return 1;
}


/**
 * Test if file is OK
 *
 * @param filepath              Filepath
 * @return                      CTRUE if file is there, CFALSE otherwise
 * @internal
 */
int
mntd_file_is_file(char *filepath)
{
	struct stat st;
	
	if (stat(filepath, &st) != 0) {
		return 0;
	}

	if (!S_ISREG(st.st_mode)) {
		return 0;
	}
	
	return 1;
}


/**
 * Get file creation timestamp
 *
 * @param filepath              Filepath
 * @return                      timestamp if successful, -1 otherwise
 * @internal
 */
int
mntd_file_get_stamp(char *filepath)
{
	struct stat st;
	
	if (stat(filepath, &st) != 0) {
		return -1;
	}

	if (!S_ISREG(st.st_mode)) {
		return -1;
	}
	
	return st.st_mtime;
}


/**
 * Get file size
 *
 * @param filepath              Filepath
 * @return                      size if successful, -1 otherwise
 * @internal
 */
int
mntd_file_get_size(char *filepath)
{
	struct stat st;
	
	if (stat(filepath, &st) != 0) {
		return -1;
	}

	if (!S_ISREG(st.st_mode)) {
		return -1;
	}
	
	return st.st_size;
}


/**
 * Read data from given file
 *
 * @param filepath              Filepath
 * @return                      String if successful, NULL otherwise
 * @internal
 */
char *
mntd_file_read(char *filepath)
{
	char *data = NULL;
	FILE *fp = NULL;
	int size = 0;
	int rsize = 0;
	int retval = 0;
	
	// opening file for reading
	fp = fopen(filepath, "r");
	if (fp != NULL) {
		// getting filesize
		size = mntd_file_get_size(filepath);
		if (size != -1) {
			data = (char *) malloc(sizeof(char)*(size+1));
			if (data != NULL) {
				// reading file
				if ((rsize = fread(data, sizeof(char), size, fp)) >= 0) {
					// EOF reached ?
					if (rsize == size) {
						retval = 1; // file successfully read
					}
				}
				data[size]='\0';
			}
		}

		// closing file
		fclose(fp);
	}
	

	// freeing memory, if error
	if (retval == 0) {
		if (data != NULL) {
			free(data);
			data = NULL;
		}
		
		return NULL;
	}
	
	return data;
}


/**
 * Write data to given file
 *
 * @param filepath              Filepath
 * @param data                  data
 * @return                      1 if successful, 0 otherwise
 * @internal
 */
int
mntd_file_write(char *filepath, char *data)
{
	FILE *fp = NULL;
	int retval = 0;
	int size = strlen(data);
	int wsize = 0;
	
	// opening file for writing
	fp = fopen(filepath, "w");
	if (fp != NULL) {
		// writing to file
		if ((wsize = fwrite(data, sizeof(char), size, fp)) >= 0) {
			// EOF reached ?
			if (wsize == size) {
				retval = 1; // file successfully written
			}
		}

		// closing file
		fclose(fp);
	}
	
	return retval;
}


/**
 * remove given file
 *
 * @param filepath              Filepath
 * @return                      CTRUE if successful, CFALSE otherwise
 * @internal
 */
int
mntd_file_remove(char *filepath)
{
	if(mntd_file_is_file(filepath)) {
		if(unlink(filepath)==0)
			return 1;
	}
	
	return 0;
}


/** @} */
