/***************************************************************************
 * CVSID: $Id: mntd_hal.c,v 1.9 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_hal.c : Watching HAL events for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include <glib.h>
#include <dbus/dbus-glib.h>

#include <libhal.h>

#include "mntd_hal.h"
#include "errmanager.h"
#include "main.h"


static void mntd_hal_device_added(const char* udi);
static void mntd_hal_device_removed(const char* udi);
static void mntd_hal_device_new_capability(const char* udi, const char* capability);
static void mntd_hal_property_changed(const char* udi, const char* key);
static void mntd_hal_property_added(const char* udi, const char* key);
static void mntd_hal_property_removed(const char* udi, const char* key);
static void mntd_hal_mainloop_integration(DBusConnection* dbus_connection);


LibHalFunctions hal_functions = {mntd_hal_mainloop_integration,
											mntd_hal_device_added, 
											mntd_hal_device_removed, 
											mntd_hal_device_new_capability,
											mntd_hal_property_changed,
											mntd_hal_property_added,
											mntd_hal_property_removed };


/**
 * @defgroup DaemonHal Hal Signal Handler
 * @ingroup MntDaemon
 * @brief Handle HAL events for MNT daemon
 * @{
 */


/** Macro for terminating the program on an unrecoverable error */
#define DIE(expr) do {printf("*** [DIE] %s:%s():%d : ", __FILE__, __FUNCTION__, __LINE__); printf expr; printf("\n"); exit(1); } while(0)


/** Invoked when a device is added to the Global Device List. Simply prints
 *  a message on stderr.
 *
 *  @param  udi                 Universal Device Id
 */
static void mntd_hal_device_added(const char* udi)
{
	// add volume to manager
	g_assert(vols!=NULL);
	vols->add_volume(vols, udi);
}


/** Invoked when a device is removed from the Global Device List. Simply
 *  prints a message on stderr.
 *
 *  @param  udi                 Universal Device Id
 */
static void mntd_hal_device_removed(const char* udi)
{
	// remove volume from manager
	g_assert(vols!=NULL);
	vols->remove_volume(vols, udi);
}


/** Invoked when device in the Global Device List acquires a new capability.
 *  Prints the name of the capability to stderr.
 *
 *  @param  udi                 Universal Device Id
 *  @param  capability          Name of capability
 */
static void mntd_hal_device_new_capability(const char* udi, const char* capability)
{
	// store new volume capability to manager
	g_assert(vols!=NULL);
	vols->new_volume_capability(vols, udi, capability);
}


/** Invoked when a property of a device in the Global Device List is
 *  changed, and we have we have subscribed to changes for that device.
 *
 *  @param  udi                 Univerisal Device Id
 *  @param  key                 Key of property
 */
static void mntd_hal_property_changed(const char* udi, const char* key)
{
	// change volume property to manager
	g_assert(vols!=NULL);
	vols->change_volume_property(vols, udi, key);
}


/** Invoked when a property of a device in the Global Device List is
 *  added, and we have we have subscribed to changes for that device.
 *
 *  @param  udi                 Univerisal Device Id
 *  @param  key                 Key of property
 */
static void mntd_hal_property_added(const char* udi, const char* key)
{
	// add volume property to manager
	g_assert(vols!=NULL);
	vols->add_volume_property(vols, udi, key);
}


/** Invoked when a property of a device in the Global Device List is
 *  removed, and we have we have subscribed to changes for that device.
 *
 *  @param  udi                 Univerisal Device Id
 *  @param  key                 Key of property
 */
static void mntd_hal_property_removed(const char* udi, const char* key)
{
	// remove volume property from manager
	g_assert(vols!=NULL);
	vols->remove_volume_property(vols, udi, key);
}


/** Invoked by libhal for integration with our mainloop. We take the
 *  easy route and use link with glib for painless integrate.
 *
 *  @param  dbus_connection     D-BUS connection to integrate
 */
static void mntd_hal_mainloop_integration(DBusConnection* dbus_connection)
{
	dbus_connection_setup_with_g_main(dbus_connection, NULL);
}


/** Initialize the HAL library
 *
 */
void mntd_hal_init(void)
{
	if(hal_initialize(&hal_functions, FALSE)) {
		MSG_ERR("hal_initialize failed (no dbus-daemon-1 running ?)\n");
		exit(1);
	}
	hal_device_property_watch_all();
}


/** Destroy the HAL library
 *
 */
void mntd_hal_shutdown(void)
{
	hal_shutdown();
}


/** @} */
