/***************************************************************************
 * CVSID: $Id: mntd_volume.h,v 1.15 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_volume.h : Volume handler for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/

#ifndef __MNTD_VOLUME_H__
#define __MNTD_VOLUME_H__

#include <dbus/dbus.h>

#include "mntd_volume_types.h"

#if defined(__cplusplus)
extern "C" {
#endif


enum {
	VOLUME_TYPE=1	/**< blabla */
};

enum {
	SIGNAL_UNKNOWN=1,		/**< last signal was unknown */
	SIGNAL_MOUNTED=2,		/**< last signal was MOUNTED */
	SIGNAL_UNMOUNTED=3	/**< last signal was UNMOUNTED */
};


/** Generate new Volume Object
 *
 * @param udi                   Volume unique  identifier got from hal
 * @param base                  Basepath for mount point
 * @return                      Pointer to Volume object,
 *                              NULL if error
 */
PVOLUME new_Volume(const char *udi, const char *base);

	
/** main volume structure
 *
 */
struct VOLUME_ {
	/* variables */
	char *udi;               /**< Store udi data */
	char *device;            /**< Store device name */
	char *mntpnt;            /**< Store mount point */
	char *fstype;            /**< Store filesystem name */
	char *base;              /**< Store mount point base path */
	int last_signal;         /**< Store last sent signal */
	
	/** Constructor
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int (*init)					(PVOLUME pv);
	
	/** Destructor
	 *
	 * @param pv                    Pointer to Volume Object
	 */
	void (*destroy)			(PVOLUME pv);
	
	/** Send signal to processes listing to dbus
	 *
	 * @param pv                    Pointer to Volume Object
	 * @param what                  Type of signal, that will be send to the 
	 *                              listening clients (One of SIGNAL_UNKNOWN | 
	 *                              SIGNAL_MOUNTED | SIGNAL_UNMOUNTED)
	 * @return                      true if successful, false otherwise
	 * @internal
	 */
	int (*_send_signal)		(PVOLUME pv, int what);
	
	/** Send mounted signal to processes listing to dbus
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      true if successful, false otherwise
	 * @internal
	 */
	int (*_send_mounted)		(PVOLUME pv);
	
	/** Send unmounted signal to processes listing to dbus
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      true if successful, false otherwise
	 * @internal
	 */
	int (*_send_unmounted)	(PVOLUME pv);

	/** Set device name
	 *
	 * @param pv                    Pointer to Volume Object
	 * @param device                Name of device
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int (*set_device)			(PVOLUME pv, const char *device);
	
	/** Get device name
	 *
	 * String within device name will be copied (with strdup), so you have
	 * to free it.
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      Name of device if already set, or NULL otherwise
	 * @internal
	 */
	char *(*get_device)		(PVOLUME pv);
	
	/** Show Device Info
	 *
	 * @param pv                    Pointer to Volume Object
	 * @internal
	 */
	void (*to_string)			(PVOLUME pv);
	
	/** Set device mount point
	 *
	 * @param pv                    Pointer to Volume Object
	 * @param mntpnt                Mount point
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int (*set_mntpnt)			(PVOLUME pv, const char *mntpnt);
	
	/** Get device mount point
	 * 
	 * String within mount point will be copied (with strdup), so you have
	 * to free it.
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      mount point if already set, or NULL otherwise
	 * @internal
	 */
	char *(*get_mntpnt)		(PVOLUME pv);
	
	/** Set device mount point
	 *
	 * @param pv                    Pointer to Volume Object
	 * @param fstype                Set filesystem name
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int (*set_fstype)			(PVOLUME pv, const char *fstype);
	
	/** Get device filesystem name
	 *
	 * String within filesystem name will be copied (with strdup), so you have
	 * to free it.
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      fstype if already set, or NULL otherwise
	 * @internal
	 */
	char *(*get_fstype)		(PVOLUME pv);
	
	/** Mount device
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int (*mount)				(PVOLUME pv);
	
	/** Unmount device
	 *
	 * Unmount device, if it is mounted by mntd, or if it's mounted in the path
	 * mntd is responsible for. So be careful, what you mount manually and don't
	 * want to be unmounted by mntd. Default is "/mnt/mntd" path, so your normal
	 * devices on "/mnt" will not be touched.
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int (*umount)				(PVOLUME pv);
	
	/** Check, if volume is in mount path
	 *
	 * @param pv                    Pointer to Volume Object
	 * @return                      true if mounted, false otherwise
	 * @internal
	 */
	int (*is_in_mntpath)		(PVOLUME pv);
	
};


#if defined(__cplusplus)
}
#endif

#endif /* __MNTD_VOLUME_H__ */
