/***************************************************************************
 * CVSID: $Id: mntd_volume_config.c,v 1.4 2004/04/06 20:49:14 stefanb Exp $
 *
 * mntd_volume_config.h : Config file parser based on expat
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <glib.h>
#include <expat.h>

#include "mntd_file.h"
#include "mntd_volume_config.h"
#include "mntd_volume_manager.h"
#include "errmanager.h"


/**
 * @defgroup DaemonHelperConfig Helper for parsing config file
 * @ingroup MntDaemon
 * @brief Some helper functions for parsing the XML config file with expat
 * @{
 */


/** Maximum nesting depth */
#define MAX_DEPTH 32

/** Maximum amount of CDATA */
#define CDATA_BUF_SIZE  1024

/** Max length of property key */
#define MAX_KEY_SIZE 128

/** Possible elements the parser can process */
enum
{
    /** Not processing a known tag */
    CURELEM_UNKNOWN  = -1,

    /** Processing the root element */
    CURELEM_MNTD = 0,
	
    /** Processing a mountpath element */
    CURELEM_MOUNTPATH = 1,
};


/** Parsing Context
 */
typedef struct
{
	/** Pointer to volume manager object */
	PVOLUMEMANAGER pvm;
	
	/** Parser object */
	XML_Parser parser;
	
	/** Config file to parse */
	char *config;
	
	/** Buffer to put CDATA in */
	char cdata_buf[CDATA_BUF_SIZE];
	
	/** Current length of CDATA buffer */
	int cdata_buf_len;
	
	/** Current depth we are parsing at */
	int depth;
	
	/** Element currently being processed */
	int curelem;
	
	/** Stack of elements being processed */
	int curelem_stack[MAX_DEPTH];
	
	/** #TRUE if parsing of document have been aborted */
	int aborted;
	
} ParsingContext;


void mntd_volume_config_parsing_abort(ParsingContext* pc);
void mntd_volume_config_start(ParsingContext *pc, const char *el, const char **attr);
void mntd_volume_config_end(ParsingContext *pc, const char* el);
void mntd_volume_config_cdata(ParsingContext *pc, const char* s, int len);
int mntd_volume_config_store_base(PVOLUMEMANAGER pvm, const char *base);


/** Abort parsing of document
 *
 *  @param  pc                  Parsing context
 */
void 
mntd_volume_config_parsing_abort(ParsingContext* pc)
{
	g_assert(pc!=NULL);
	
	/* Grr, expat can't abort parsing */
	MSG_ERR("Aborting parsing of document");
	pc->aborted = TRUE;
}


/** Called by expat when an element begins.
 *
 *  @param  pc                  Parsing context
 *  @param  el                  Element name
 *  @param  attr                Attribute key/value pairs
 */
void 
mntd_volume_config_start(ParsingContext *pc, const char *el, const char **attr)
{
	int i = 0;
	
	g_assert(pc!=NULL);
	g_assert(el!=NULL);
	g_assert(attr!=NULL);
	
	MSG_DEBUG("start: %s", el);
	
	for (i=0; attr[i]; i+=2) {
		MSG_DEBUG("    %s='%s'", attr[i], attr[i+1]);
	}
	
	// check abort flag
	if(pc->aborted) {
		return;
	}
	
	// set cdata buffer start
	pc->cdata_buf_len = 0;
	
	if (strcmp(el, "mntd") == 0) {
		if( pc->curelem!=CURELEM_UNKNOWN ) {
			MSG_ERR("%s:%d:%d: Element <mntd> must be a top-level "
							"element", 
							pc->config, 
							XML_GetCurrentLineNumber(pc->parser), 
							XML_GetCurrentColumnNumber(pc->parser));
			mntd_volume_config_parsing_abort(pc);
		}
		pc->curelem = CURELEM_MNTD;
		// TODO: handle elements data
		
	} else if (strcmp(el, "mountpath") == 0) {
		if( pc->curelem!=CURELEM_MNTD ) {
			MSG_ERR("%s:%d:%d: Element <mountpath> can only be inside <mntd>",
							pc->config, 
							XML_GetCurrentLineNumber(pc->parser), 
							XML_GetCurrentColumnNumber(pc->parser));
			mntd_volume_config_parsing_abort(pc);
		}
		pc->curelem = CURELEM_MOUNTPATH;
		// TODO: handle elements data
		
	} else {
		MSG_ERR("%s:%d:%d: Unknown element <%s>",
						pc->config, 
						XML_GetCurrentLineNumber(pc->parser), 
						XML_GetCurrentColumnNumber(pc->parser), el);
		mntd_volume_config_parsing_abort(pc);
	}
	
	// Nasty hack
	g_assert((pc->depth)<MAX_DEPTH);
	
	// increment depth counter
	pc->depth++;        
	
	// store depth
	pc->curelem_stack[pc->depth] = pc->curelem;
	
	return;
}


/** Called by expat when an element ends.
 *
 *  @param  pc                  Parsing context
 *  @param  el                  Element name
 */
void 
mntd_volume_config_end(ParsingContext *pc, const char* el)
{
	g_assert(pc!=NULL);
	g_assert(el!=NULL);
	
	MSG_DEBUG("end: %s", el);
	
	// check abort flag
	if (pc->aborted) {
		return;
	}
	
	// finalize cdata buffer
	pc->cdata_buf[pc->cdata_buf_len]='\0';
	
	// TODO: handle data
	if (pc->curelem==CURELEM_MOUNTPATH) {
		MSG_DEBUG("content for mountpath='%s'", pc->cdata_buf);
		if(mntd_volume_config_store_base(pc->pvm, pc->cdata_buf) == -1) {
			MSG_ERR("Could not store new mountpath");
			mntd_volume_config_parsing_abort(pc);
		}
	}
	
	// reset cdata offset
	pc->cdata_buf_len = 0;
	
	// decrement depth counter
	pc->depth--;
	
	// maintain curelem
	pc->curelem = pc->curelem_stack[pc->depth];
	
	return;
}


/** Called when there is CDATA 
 *
 *  @param  pc                  Parsing context
 *  @param  s                   Pointer to data
 *  @param  len                 Length of data
 */
void 
mntd_volume_config_cdata(ParsingContext *pc, const char* s, int len)
{
	int bytes_left;
	int bytes_to_copy;
	
	g_assert(pc!=NULL);
	g_assert(s!=NULL);
	
	// handle abort flag
	if (pc->aborted) {
		return;
	}
	
	bytes_left = CDATA_BUF_SIZE - pc->cdata_buf_len;
	if (len>bytes_left) {
		MSG_ERR("CDATA in element larger than %d", CDATA_BUF_SIZE);
	}
	
	bytes_to_copy = len;
	if (bytes_to_copy>bytes_left) {
		bytes_to_copy = bytes_left;
	}
	
	if (bytes_to_copy>0) {
		memcpy(pc->cdata_buf + pc->cdata_buf_len, s, bytes_to_copy);
	}
	
	pc->cdata_buf_len += bytes_to_copy;
	
	return;
}


/** Parse the config file
 *
 * @param config                Path of configfile
 * @param pvm                   Pointer to Volume Manager Object
 * @return                      0 if successful, -1 otherwise
 */
int 
mntd_volume_config_parse(char *config, PVOLUMEMANAGER pvm)
{
	int res = 0;
	char *data = NULL;
	XML_Parser parser = NULL;
	int size = 0;
	ParsingContext *parsing_context = NULL;
	
	g_assert(pvm!=NULL);
	
	// reading configuration file
	MSG_INF("Loading configuration from '%s'.", config);
	if(!mntd_file_is_file(config)) {
		MSG_EMERG("Couldn't read config file from '%s'", config);
		return -1;
	}
	
	if((size = mntd_file_get_size(config)) == -1 ) {
		MSG_EMERG("Couldn't get config file size from '%s'", config);
		goto error;
	}
	
	// read config file data
	data = mntd_file_read(config);
	if (data == NULL) {
		MSG_EMERG("Couldn't read config file from '%s'", config);
		goto error;
	}
	
	// read xml data
	parser = XML_ParserCreate(NULL);
	
	// initialize parsing context
	parsing_context = (ParsingContext*) malloc(sizeof(ParsingContext));
	if (parsing_context == NULL) {
		MSG_EMERG("Out of memory.");
		goto error;
	}
	parsing_context->depth = 0;
	parsing_context->cdata_buf_len = 0;
	memset(parsing_context->cdata_buf, 0, sizeof(parsing_context->cdata_buf));
	parsing_context->curelem = CURELEM_UNKNOWN;
	parsing_context->aborted = FALSE;
	parsing_context->parser = parser;
	parsing_context->config = config;
	parsing_context->pvm = pvm;
	
	// parse xml data
	XML_SetElementHandler(parser, 
								(XML_StartElementHandler) mntd_volume_config_start, 
								(XML_EndElementHandler) mntd_volume_config_end);
	XML_SetCharacterDataHandler(parser, 
								(XML_CharacterDataHandler) mntd_volume_config_cdata);
	XML_SetUserData(parser, parsing_context);
	if (XML_Parse(parser, data, size, 1) == XML_STATUS_ERROR) {
		MSG_EMERG("Parse error at line %d:\n%s\n",
												XML_GetCurrentLineNumber(parser),
												XML_ErrorString(XML_GetErrorCode(parser)));
	}
	
	if (parsing_context->aborted == TRUE) {
		goto error;
	}
	
	goto cleanup;

error:
	res = -1;
	
cleanup:
	// free parsing context
	if (parsing_context != NULL) {
		// TODO: store parsing context here
		free(parsing_context);
		parsing_context = NULL;
	}
	
	// free config file data
	if (data != NULL) {
		free(data);
		data = NULL;
	}
	
	return res;
}


/** Store base mountpath from config file
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param base                  base mountpath
 * @return                      0 if successful, -1 otherwise
 */
int 
mntd_volume_config_store_base(PVOLUMEMANAGER pvm, const char *base)
{
	int res = 0;
	char *tmp = NULL;
	
	g_assert(pvm!=NULL);
	
	// free already stored base path
	if (pvm->base != NULL) {
		free(pvm->base);
		pvm->base = NULL;
	}
	
	// store new value
	if (pvm->base == NULL) {
		tmp = strdup(base);
		if (tmp == NULL) {
			goto error;
		}
		pvm->base = tmp;
	} else {
		goto error;
	}

	goto cleanup;

error:
	if (tmp!=NULL) {
		free(tmp);
		tmp = NULL;
	}
	res = -1;

cleanup:
	
	return res;
}


/** @} */
