/***************************************************************************
 * CVSID: $Id: mntd_volume_hash.c,v 1.10 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_volume_hash.c : Volume Manager hash functions for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "mntd_volume_hash.h"
#include "errmanager.h"


/* private prototypes */
PHASHMAPELEMENT new_HashMapElement(PHASHMAP phm, void (*func_free)(void *data), 
												const char *key, void *userdata);
void hashmapelement_destroy(PHASHMAPELEMENT phme);
int hashmap_init(PHASHMAP phm);
void hashmap_destroy(PHASHMAP phm);
int hashmap_hash_element(const void *obj);
int hashmap_compare_element(const void *obj1, const void *obj2);
void hashmap_free_element(void *userdata);

/* public prototypes */
int hashmap_add(PHASHMAP phm, const char *key, void *data);
int hashmap_remove(PHASHMAP phm, const char *key);
int hashmap_size(PHASHMAP phm);
int hashmap_contains(PHASHMAP phm, const char *key);
int hashmap_get(PHASHMAP phm, const char *key, void **userdata);
int hashmap_update(PHASHMAP phm, const char *key, void *userdata);
int hashmap_foreach_call_function(PHASHMAP phm, int (*func)(void *data, void *userdata), void *userdata);


/** HashMapElement Object Definition
 *
 * @param phm                   Pointer to HashMap object
 * @param func_free             Pointer to callback function, that will be 
 *                              called each time an element will be removed
 * @param key                   Key for this hashmap element
 * @param userdata              Userdata pointer, that will be passed to callback function
 * @return                      Pointer to new HashMapElement object,
 *                              NULL if error
 * @internal
 */
PHASHMAPELEMENT 
new_HashMapElement(PHASHMAP phm, void (*func_free)(void *userdata), const char *key, void *userdata)
{
	PHASHMAPELEMENT phme = NULL;
	
	g_assert(phm!=NULL);
	g_assert(func_free!=NULL);
	g_assert(key!=NULL);
	
	//MSG_DEBUG("new_HashMapElement('%s', %p) called", key, userdata);
	
	phme = (PHASHMAPELEMENT) malloc(sizeof(HASHMAPELEMENT));
	if (phme == NULL) {
		MSG_ERR("Out of memory");
		return NULL;
	}
	
	// clear structure
	memset(phme, 0, sizeof(HASHMAPELEMENT));
	
	// initialize data
	phme->phm = phm;
	phme->key = strdup(key);
	phme->userdata = userdata;
	
	// private methods
	phme->destroy = hashmapelement_destroy;
	phme->free = func_free;
	
	//MSG_DEBUG("new_HashMapElement('%s', %p) returned %p", key, userdata, phme);
	
	// return object
	return phme;
}


/** HashMapElement Destructor
 *
 * @param phme                  Pointer to HashMapElement Object
 * @internal
 */
void
hashmapelement_destroy(PHASHMAPELEMENT phme)
{
	g_assert(phme!=NULL);
	
	//MSG_DEBUG("hashmapelement_destroy(%p) called (key='%s', userdata=%p)", phme, phme->key, phme->userdata);

	// free memory for key
	if (phme->key != NULL) {
		free(phme->key);
		phme->key = NULL;
	}
	
	// call user defined functions
	if ((phme->free != NULL) && (phme->userdata != NULL)) {
		phme->free(phme->userdata);
		phme->free = NULL;
		phme->userdata = NULL;
	}
	
	// clear structure
	memset(phme, 0, sizeof(HASHMAPELEMENT));
	free(phme);
	phme = NULL;
	
	return;
}


PHASHMAP 
new_HashMap(unsigned int bucket, 
	void (*func_free)(void *userdata)
	)
{
	PHASHMAP phm = NULL;
	
	phm = (PHASHMAP) malloc(sizeof(HASHMAP));
	if (phm == NULL) {
		MSG_ERR("Out of memory");
		return NULL;
	}
	
	// clear structure
	memset(phm, 0, sizeof(HASHMAP));
	
	// no. of buckets (should be prime num)
	if (bucket>0) {
		phm->bucket = bucket;
	} else {
		MSG_INF("bucket not correctly setup. Using %d as default.", DEFAULT_BUCKETS);
		phm->bucket = DEFAULT_BUCKETS;
	}

	// private methods
	phm->init = hashmap_init;
	phm->destroy = hashmap_destroy;
	// hash, compare, free methods (implemented for each object)
	phm->hash = hashmap_hash_element;
	phm->compare = hashmap_compare_element;
	if (func_free != NULL) {
		phm->free = func_free;
	} else {
		phm->free = hashmap_free_element;
	}
	// public methods
	phm->add = hashmap_add;
	phm->remove = hashmap_remove;
	phm->size = hashmap_size;
	phm->contains = hashmap_contains;
	phm->get = hashmap_get;
	phm->update = hashmap_update;
	phm->foreach = hashmap_foreach_call_function;
	
	// call constructor
	if(phm->init(phm) == -1) {
		free(phm);
		phm = NULL;
	}

	// return object
	return phm;
}


/** HashMap Constructor
 *
 * @param phm                   Pointer to HashMap Object
 * @return                      0 if successful, -1 otherwise
 * @internal
 */
int
hashmap_init(PHASHMAP phm)
{
	CHTbl *htbl = NULL;

	if (phm==NULL) {
		return -1;
	}
	
	// get mem for structure
	htbl = (CHTbl *) malloc(sizeof(CHTbl));
	if (htbl == NULL) {
		MSG_ERR("not enough mem for hash table");
		return -1;
	}
	
	// initialize table
	if (chtbl_init(htbl, phm->bucket, phm->hash, phm->compare, phm->free) != 0) {
		MSG_ERR("could not initialize hash table");
		if (htbl!=NULL) free(htbl);
		return -1;
	}
	
	// store table pointer
	phm->hm = htbl;
	
	// return success
	return 0;
}


/** HashMap Destructor
 *
 * Call it to free to hashmap
 *
 * @param phm                   Pointer to HashMap Object
 */
void
hashmap_destroy(PHASHMAP phm)
{
	if ((phm==NULL) || ((phm->hm)==NULL)) {
		return;
	}
	
	chtbl_destroy(phm->hm);
	free(phm->hm);
	phm->hm = NULL;
	return;
}


/** HashMap hash for element
 *
 * @param key                   Key for hash entry
 * @return                      hash from key (buckets will be handled by chtbl)
 * @internal
 */
int
hashmap_hash_element(const void *obj)
{
	char *key = NULL;
	unsigned int crc = 0;
	
	if (obj==NULL) {
		MSG_EMERG("Hashmap HASH function called with wrong parameter ! (obj=0x%08x)", obj);
		return 0;
	}
	
	key = ((PHASHMAPELEMENT)obj)->key;
	while(*key!='\0') {
		crc += *key++;
	}
	
	return crc;
}


/** HashMap compare for element
 *
 * @param obj1                  Key for hash entry (default: element)
 * @param obj2                  Key for hash entry (default: element)
 * @return                      compare status
 * @internal
 */
int
hashmap_compare_element(const void *obj1, const void *obj2)
{
	char *key1 = NULL;
	char *key2 = NULL;
	int res = 0;
	
	if ((obj1==NULL) || (obj2==NULL)) {
		MSG_EMERG("Hashmap MATCH function called with wrong parameters ! (obj1=0x%08x, obj2=0x%08x)", obj1, obj2);
		return 0;
	}
		
	if (obj1==obj2) {
		MSG_EMERG("Hashmap MATCH function called with identical parameters ! (obj1=obj2=0x%08x)", obj1);
		return 0;
	}
		
	key1 = ((PHASHMAPELEMENT)obj1)->key;
	key2 = ((PHASHMAPELEMENT)obj2)->key;

	res = strcmp(key1, key2);

	return res == 0 ? 1 : 0;
}


/** HashMap hash for element
 *
 * @param data                  Pointer to data (default: element)
 * @internal
 */
void
hashmap_free_element(void *userdata)
{
	if (userdata != NULL) {
		MSG_WARNING("You really should set the 'free()' method !");
	}
	return;
}


/** Add to HashMap
 *
 * @param phm                   Pointer to HashMap Object
 * @param key                   Key for hash entry
 * @param data                  Data pointer to store
 * @return                      0 if successfully inserted
 *                             -1 if error while inserting
 *                              1 if element is already in list
 */
int
hashmap_add(PHASHMAP phm, const char *key, void *data)
{
	PHASHMAPELEMENT phme = NULL;
	int retval;
	
	if ((phm==NULL) || (key==NULL)) {
		return -1;
	}

	phme = new_HashMapElement(phm, phm->free, key, data);
	if (phme == NULL) {
		return -1;
	}
	
	retval = chtbl_insert(phm->hm, phme);
	
	return retval;
}


/** Remove from HashMap
 *
 * @param phm                   Pointer to HashMap Object
 * @param key                   Key for hash entry
 * @return                      0 if successful, -1 otherwise
 */
int
hashmap_remove(PHASHMAP phm, const char *key)
{
	PHASHMAPELEMENT tmp = NULL;
	PHASHMAPELEMENT phme = NULL;
	void *chtbl_data = NULL;
	int retval;
	
	if ((phm==NULL) || ((phm->hm)==NULL) || (key==NULL)) {
		return -1;
	}

	tmp = new_HashMapElement(phm, phm->free, key, NULL);
	if (tmp == NULL) {
		return -1;
	}
	
	chtbl_data = (void *)tmp;
	retval = chtbl_remove(phm->hm, &chtbl_data);
	if (retval == 0) {
		if (chtbl_data != NULL) {
			phme = (PHASHMAPELEMENT)chtbl_data;
			if (phme != tmp) {
				if (phme != NULL) {
					phme->destroy(phme);
				}
			}
		}
	}
	tmp->destroy(tmp); tmp=NULL;
	
	return retval;
}


/** Size of HashMap
 *
 * @param phm                   Pointer to HashMap Object
 * @return                      >=0 if successful, -1 otherwise
 */
int
hashmap_size(PHASHMAP phm)
{
	if ((phm==NULL) || ((phm->hm)==NULL)) {
		return -1;
	}
	
	return chtbl_size(phm->hm);
}


/** HashMap contains
 *
 * @param phm                   Pointer to HashMap Object
 * @param key                   Key to search for
 * @return                      1 if found, 0 otherwise
 */
int
hashmap_contains(PHASHMAP phm, const char *key)
{
	PHASHMAPELEMENT tmp = NULL;
	void *data = NULL;
	int retval;
	
	if ((phm==NULL) || ((phm->hm)==NULL)) {
		return 0;
	}
	
	tmp = new_HashMapElement(phm, phm->free, key, NULL);
	if (tmp == NULL) {
		return 0;
	}
	
	data = (void *)tmp;
	retval = chtbl_lookup(phm->hm, &data);
	tmp->destroy(tmp); tmp=NULL;
	
	return (retval==0) ? 1 : 0;
}


/** Get userdata from HashMap
 *
 * @param phm                   Pointer to HashMap Object
 * @param key                   Key for hash entry
 * @param userdata              return value
 * @return                      0 if successful, -1 otherwise
 */
int
hashmap_get(PHASHMAP phm, const char *key, void **userdata)
{
	PHASHMAPELEMENT tmp = NULL;
	void *data = NULL;
	int retval;
	
	if ((phm==NULL) || ((phm->hm)==NULL)) {
		return -1;
	}
	
	tmp = new_HashMapElement(phm, phm->free, key, NULL);
	if (tmp == NULL) {
		return -1;
	}
	
	data = (void *)tmp;
	retval = chtbl_lookup(phm->hm, &data);
	tmp->destroy(tmp); tmp=NULL;
	
	if (retval != 0) {
		*userdata = NULL;
		return -1;
	}
	
	// return data
	*userdata = ((PHASHMAPELEMENT)data)->userdata;
	return 0;
}


/** Update element in HashMap
 *
 * @param phm                   Pointer to HashMap Object
 * @param key                   Key for hash entry
 * @param data                  Data pointer to update
 * @return                      0 if successful, -1 otherwise
 */
int
hashmap_update(PHASHMAP phm, const char *key, void *userdata)
{
	PHASHMAPELEMENT tmp = NULL;
	PHASHMAPELEMENT phme = NULL;
	void *data = NULL;
	int retval;
	
	if ((phm==NULL) || (key==NULL)) {
		return -1;
	}

	if (phm->contains(phm, key)) {
		tmp = new_HashMapElement(phm, phm->free, key, NULL);
		if (tmp == NULL) {
			return -1;
		}
	
		data = (void *)tmp;
		retval = chtbl_lookup(phm->hm, &data);
		tmp->destroy(tmp); tmp=NULL;
		
		if (retval == 0) {
			phme = (PHASHMAPELEMENT)data;
			if (phme != NULL) {
				phme->free(phme->userdata);
				phme->userdata = userdata;
			}
		}
	}
	
	return -1;
}


/** Iterate over HashMap
 *
 * @param phm                   Pointer to HashMap Object
 * @param func                  callback function
 * @param data                  User data pointer
 * @return                      0 if successful, -1 otherwise
 */
int
hashmap_foreach_call_function(PHASHMAP phm, 
	int (*func)(void *data, void *userdata), 
   void *userdata)
{
	int i=0;
	
	if ((phm==NULL) || (func==NULL) || (phm->hm==NULL)) {
		return -1;
	}

	for (i=0; i<phm->hm->buckets; i++) {
		list_for_each_call(&phm->hm->table[i], func, userdata);
	}
	
	return 0;
}
