/***************************************************************************
 * CVSID: $Id: mntd_volume_hash.h,v 1.8 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_volume_hash.h : Volume Manager hash functions for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/

#ifndef __HASHMAP_H__
#define __HASHMAP_H__

#include "mntd_volume_hash_types.h"
#include "hashmap.h"
#include "linklist.h"

#if defined(__cplusplus)
extern "C" {
#endif

	
/**
 * HashMap wrapper
 *
 * @defgroup HashMap Hash Map Wrapper
 * @author Stefan Bambach (stefan@bambach.biz)
 * @version 0.1
 * @brief Hash Map
 * @{
 */


#define DEFAULT_BUCKETS 1021    /**< hash table bucket define */

	
/** Create new HashMap object
 *
 * Call the destroy() method to remove hashmap
 *
 * @param bucket                Number of buckets, that this hashmap will use
 * @param func_free             Pointer to callback function, that will be 
 *                              called each time an element will be removed
 * @return                      Pointer to HashMap object,
 *                              NULL if error
 */
PHASHMAP new_HashMap(unsigned int bucket, void (*func_free)(void *obj));

	
/** main hash map element structure
 *
 */
struct HASHMAPELEMENT_ {
	PHASHMAP phm;                /**< Back pointer to hashmap object */
	char *key;                   /**< Elements key */
	void *userdata;              /**< Elements userdata pointer */
	
	/** HashMapElement Destructor
	 *
	 * @param phme                  Pointer to HashMapElement Object
	 * @internal
	 */
	void	(*destroy)		(PHASHMAPELEMENT phme);
	
	/** Free callback pointer for element
	 *
	 * @param data               Pointer to data, to be freed
	 */
	void	(*free)			(void *data);
};


/** main hash map structure
 *
 */
struct HASHMAP_ {
	CHTbl *hm;                   /**< Pointer to chained hashmap table  */
	unsigned int bucket;         /**< Number of buckets to be used */
	
	/** HashMap Constructor
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int	(*init)			(PHASHMAP phm);
	
	/** HashMap hash for element
	 *
	 * @param key                   Key for hash entry
	 * @return                      hash from key (buckets will be handled by chtbl)
	 * @internal
	 */
	int	(*hash)			(const void *obj);
	
	/** HashMap compare for element
	 *
	 * @param obj1                  Key for hash entry (default: element)
	 * @param obj2                  Key for hash entry (default: element)
	 * @return                      compare status
	 * @internal
	 */
	int	(*compare)		(const void *obj1, const void *obj2);
	
	/** HashMap hash for element
	 *
	 * @param data                  Pointer to data (default: element)
	 * @internal
	 */
	void	(*free)			(void *data);

	/** HashMap Destructor
	 *
	 * Call it to free to hashmap
	 *
	 * @param phm                   Pointer to HashMap Object
	 */
	void	(*destroy)		(PHASHMAP phm);
	
	/** Add to HashMap
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @param key                   Key for hash entry
	 * @param data                  Data pointer to store
	 * @return                      0 if successfully inserted
	 *                             -1 if error while inserting
	 *                              1 if element is already in list
	 */
	int	(*add)			(PHASHMAP phm, const char *key, void *data);
	
	/** Remove from HashMap
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @param key                   Key for hash entry
	 * @return                      0 if successful, -1 otherwise
	 */
	int	(*remove)		(PHASHMAP phm, const char *key);
	
	/** Size of HashMap
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @return                      >=0 if successful, -1 otherwise
	 */
	int	(*size)			(PHASHMAP phm);
	
	/** HashMap contains
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @param key                   Key to search for
	 * @return                      1 if found, 0 otherwise
	 */
	int	(*contains)		(PHASHMAP phm, const char *key);
	
	/** Get userdata from HashMap
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @param key                   Key for hash entry
	 * @param userdata              return value
	 * @return                      0 if successful, -1 otherwise
	 */
	int	(*get)			(PHASHMAP phm, const char *key, void **userdata);
	
	/** Update element in HashMap
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @param key                   Key for hash entry
	 * @param data                  Data pointer to update
	 * @return                      0 if successful, -1 otherwise
	 */
	int	(*update)		(PHASHMAP phm, const char *key, void *userdata);
	
	/** Iterate over HashMap
	 *
	 * @param phm                   Pointer to HashMap Object
	 * @param func                  callback function
	 * @param data                  User data pointer
	 * @return                      0 if successful, -1 otherwise
	 */
	int	(*foreach)		(PHASHMAP phm, int (*func)(void *data, void *userdata), void *userdata);
	
};


/** @} */


#if defined(__cplusplus)
}
#endif

#endif /* __HASHMAP_H__ */
