/***************************************************************************
 * CVSID: $Id: mntd_volume_manager.c,v 1.27 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_volume_manager.c : Volume Manager handler for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include <libhal.h>
#include <libmnt/libmnt.h>   /* for common defines etc. */

#include "mntd_volume_manager_helper.h"
#include "mntd_volume_manager.h"
#include "mntd_volume_hash.h"
#include "mntd_volume_config.h"
#include "mntd_volume.h"
#include "mntd_hal.h"
#include "mntd_file.h"
#include "errmanager.h"


/* Private Methods */
int mntd_volume_manager_init(PVOLUMEMANAGER pvm);
void mntd_volume_manager_destroy(PVOLUMEMANAGER pvm);
void mntd_volume_manager_func_free(void *userdata);
void mntd_volume_manager_add_volume(PVOLUMEMANAGER pvm, const char *udi);
void mntd_volume_manager_remove_volume(PVOLUMEMANAGER pvm, const char *udi);
void mntd_volume_manager_new_volume_capability(PVOLUMEMANAGER pvm, const char *udi, const char *capability);
void mntd_volume_manager_add_volume_property(PVOLUMEMANAGER pvm, const char *udi, const char *property);
void mntd_volume_manager_remove_volume_property(PVOLUMEMANAGER pvm, const char *udi, const char *property);
void mntd_volume_manager_change_volume_property(PVOLUMEMANAGER pvm, const char *udi, const char *property);
void mntd_volume_manager_rescan(PVOLUMEMANAGER pvm);
void mntd_volume_manager_foreach_mounted(PVOLUMEMANAGER pvm, 
												int (*func)(void *data, void *userdata), 
												void *userdata);
char *mntd_volume_manager_get_mntpnt(PVOLUMEMANAGER pvm, const char *udi);
int mntd_volume_manager_contains(PVOLUMEMANAGER pvm, const char *udi);


/**
 * @defgroup DaemonVolumeManager Volume Manager Handler
 * @ingroup MntDaemon
 * @brief Manage Volumes in maps
 * @{
 */


/** Generate new Volume Manager Object
 *
 * @param config                Path to configfile
 * @return                      Pointer to Volume Manager object,
 *                              NULL if error
 */
PVOLUMEMANAGER new_VolumeManager(const char *config)
{
	PVOLUMEMANAGER pvm = NULL;
	PHASHMAP phm = NULL;
	char *s_config = NULL;
	
	// check for error
	g_assert(config!=NULL);
	
	// get mem for volume object
	pvm = (PVOLUMEMANAGER) malloc(sizeof(VOLUMEMANAGER));
	if (pvm == NULL) {
		return NULL;
	}
	
	// clear structure memory
	memset(pvm, 0, sizeof(VOLUMEMANAGER));
	
	// generate hashmap
	phm = new_HashMap(DEFAULT_BUCKETS, mntd_volume_manager_func_free);
	if (phm == NULL) {
		goto error;
	}
	pvm->phm = phm;
	
	// setting config file
	s_config = strdup(config);
	if (s_config == NULL) {
		goto error;
	}
	pvm->config = s_config;
	
	// private methods
	pvm->init = mntd_volume_manager_init;
	
	// public methods
	pvm->destroy = mntd_volume_manager_destroy;
	pvm->add_volume = mntd_volume_manager_add_volume;
	pvm->remove_volume = mntd_volume_manager_remove_volume;
	pvm->new_volume_capability = mntd_volume_manager_new_volume_capability;
	pvm->add_volume_property = mntd_volume_manager_add_volume_property;
	pvm->remove_volume_property = mntd_volume_manager_remove_volume_property;
	pvm->change_volume_property = mntd_volume_manager_change_volume_property;
	pvm->rescan = mntd_volume_manager_rescan;
	pvm->foreach_mounted = mntd_volume_manager_foreach_mounted;
	pvm->get_mntpnt = mntd_volume_manager_get_mntpnt;
	pvm->contains = mntd_volume_manager_contains;
	
	// call constructor
	if(pvm->init(pvm) == -1) {
		goto error;
	}
	
	goto exit;
	
error:
	pvm->destroy(pvm);
	pvm = NULL;
	
exit:
	// return object
	return pvm;
}


/** Constructor
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @return                      0 if successful, -1 otherwise
 */
int mntd_volume_manager_init(PVOLUMEMANAGER pvm)
{
	g_assert(pvm!=NULL);
	
	return mntd_volume_config_parse(pvm->config, pvm);
}


/** Destructor
 *
 * @param pvm                   Pointer to Volume Manager Object
 */
void mntd_volume_manager_destroy(PVOLUMEMANAGER pvm)
{
	PHASHMAP phm = NULL;
	
	g_assert(pvm!=NULL);
	
	// destroy base path name
	if (pvm->base != NULL) {
		free(pvm->base);
		pvm->base = NULL;
	}

	// destroy config file name
	if (pvm->config != NULL) {
		free(pvm->config);
		pvm->config = NULL;
	}

	// destroy hash
	phm = pvm->phm;
	if (phm != NULL) {
		phm->destroy(phm);
	}
	pvm->phm = NULL;
	
	// free volume manager structure
	memset(pvm, 0, sizeof(VOLUMEMANAGER));
	free(pvm);
	
	return;
}


/** Callback function for freeing memory for Volume object
 *
 * @param userdata              Pointer userdata
 */
void mntd_volume_manager_func_free(void *userdata)
{
	PVOLUME pv = NULL;
	
	if (userdata != NULL) {
		
		//MSG_DEBUG("----------==========  DESTROYING VOLUME OBJECT  ==========----------");
		
		pv = (PVOLUME) userdata;
		pv->destroy(pv);
		pv = NULL;
	} else {
		//MSG_ERR("No volume object to be freed given");
	}
	
	return;
}


/** Add volume to manager
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 */
void mntd_volume_manager_add_volume(PVOLUMEMANAGER pvm, const char *udi)
{
	PVOLUME pv = NULL;
	PHASHMAP phm = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	if (vmh_hal_device_is_volume(udi)) {
		
		//MSG_DEBUG("----------==========  ADD VOLUME  ==========---------- (%s)", udi);

		// generate new volume object
		pv = new_Volume(udi, pvm->base);
		
		if (pv!=NULL) {
			if (phm->add(phm, udi, pv) == 0) {
				vmh_try_n_set_volumes_data(pv, udi);
			} else {
				pv->destroy(pv);
				pv = NULL;
				//MSG_ERR("Couldn't add Volume (%s) to Manager", udi);
			}
		}
	}
	
	return;
}


/** Remove volume from manager
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 */
void mntd_volume_manager_remove_volume(PVOLUMEMANAGER pvm, const char *udi)
{
	PHASHMAP phm = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	
	phm = pvm->phm;
	if (phm!=NULL) {
		if (phm->contains(phm, udi)) {
			
			//MSG_DEBUG("----------==========  REMOVE VOLUME  ==========---------- (%s)", udi);
		
			phm->remove(phm, udi);
		}
	} else {
		//MSG_ERR("Hashmap pointer is NULL -> Cannot remove volume");
	}
	
	return;
}


/** Add volume capability to volume
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 * @param capability            Pointer to new capability
 */
void mntd_volume_manager_new_volume_capability(PVOLUMEMANAGER pvm, const char *udi, const char *capability)
{
	PHASHMAP phm = NULL;
	PVOLUME pv = NULL;
	void *data = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	if (phm->get(phm, udi, &data) == 0) {
		
		//MSG_DEBUG("----------==========  NEW VOLUME CAPABILITY  ==========---------- (%s:%s)", udi, capability);

		pv = (PVOLUME) data;
		
		if (pv!=NULL) {
			vmh_try_n_set_volumes_data_by_property(pv, udi, capability);
		}
	}
	
	return;
}


/** Add volume property
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 * @param property              property name of removed volume
 */
void mntd_volume_manager_add_volume_property(PVOLUMEMANAGER pvm, const char *udi, const char *property)
{
	PHASHMAP phm = NULL;
	PVOLUME pv = NULL;
	void *data = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	if (phm->get(phm, udi, &data) == 0) {
		
		//MSG_DEBUG("----------==========  ADD VOLUME PROPERTY  ==========---------- (%s:%s)", udi, property);

		pv = (PVOLUME) data;
		
		if (pv!=NULL) {
			vmh_try_n_set_volumes_data_by_property(pv, udi, property);
		}
	}
	
	return;
}


/** Remove volume property
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 * @param property              property name of removed volume
 */
void mntd_volume_manager_remove_volume_property(PVOLUMEMANAGER pvm, const char *udi, const char *property)
{
	PHASHMAP phm = NULL;
	PVOLUME pv = NULL;
	void *data = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	if (phm->get(phm, udi, &data) == 0) {
		
		//MSG_DEBUG("----------==========  REMOVE VOLUME PROPERTY  ==========---------- (%s:%s)", udi, property);

		pv = (PVOLUME) data;
		
		if (pv!=NULL) {
			vmh_try_n_set_volumes_data_by_property(pv, udi, property);
		}
	}
	
	return;
}


/** Change volume property
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 * @param property              property name of removed volume
 */
void mntd_volume_manager_change_volume_property(PVOLUMEMANAGER pvm, const char *udi, const char *property)
{
	PHASHMAP phm = NULL;
	PVOLUME pv = NULL;
	void *data = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	if (phm->get(phm, udi, &data) == 0) {
		
		//MSG_DEBUG("----------==========  CHANGE VOLUME PROPERTY  ==========---------- (%s:%s)", udi, property);

		pv = (PVOLUME) data;
		
		if (pv != NULL) {
			vmh_try_n_set_volumes_data_by_property(pv, udi, property);
		}
	}
	
	return;
}


/** Rescan HAL devices
 *
 * @param pvm                   Pointer to Volume Manager Object
 */
void mntd_volume_manager_rescan(PVOLUMEMANAGER pvm)
{
	int i = 0;
	int num_devices = 0;
	char **device_names = NULL;
	char *udi = NULL;
	
	g_assert(pvm!=NULL);
	
	device_names = hal_get_all_devices(&num_devices);
	
	if (device_names == NULL) {
		MSG_ERR("Couldn't get devices from HAL (hald not running ?).");
		return;
	}
	
	for (i=0; i<num_devices; i++) {
		udi = device_names[i];
		if(vmh_hal_device_is_safe(udi)) {
			if (vmh_hal_device_is_volume(udi)) {
				pvm->add_volume(pvm, udi);
			}
		}
	}
	
	return;
}


/** call function for each mounted volume
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param func                  Pointer to callback function
 * @param userdata              userdata pointer that will be passed to callback function
 */
void mntd_volume_manager_foreach_mounted(PVOLUMEMANAGER pvm, 
												int (*func)(void *data, void *userdata), 
												void *userdata)
{
	PHASHMAP phm = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(func!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	phm->foreach(phm, func, userdata);
	
	return;
}


/** Get specific mount point
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 * @return                      Mount point, or NULL if error
 */
char *mntd_volume_manager_get_mntpnt(PVOLUMEMANAGER pvm, const char *udi)
{
	PHASHMAP phm = NULL;
	PVOLUME pv = NULL;
	void *data = NULL;
	char *mntpnt = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	if (phm->get(phm, udi, &data) == 0) {
		pv = (PVOLUME) data;
		if (pv!=NULL) {
			// check for mntd handled mntpnt
			if (pv->is_in_mntpath(pv)) {
				mntpnt = pv->get_mntpnt(pv);
				if (mntpnt!=NULL) {
					return mntpnt;
				}
			}
		}
	}
	
	return NULL;
}


/** Check if given volume exists
 *
 * @param pvm                   Pointer to Volume Manager Object
 * @param udi                   udi (key) of object
 * @return                      1 if found, 0 otherwise
 */
int mntd_volume_manager_contains(PVOLUMEMANAGER pvm, const char *udi)
{
	PHASHMAP phm = NULL;
	
	g_assert(pvm!=NULL);
	g_assert(udi!=NULL);
	g_assert((pvm->phm)!=NULL);
	phm = pvm->phm;
	
	return phm->contains(phm, udi);
}


/** @} */
