/***************************************************************************
 * CVSID: $Id: mntd_volume_manager.h,v 1.14 2004/05/26 19:12:40 stefanb Exp $
 *
 * mntd_volume_manager.h : Volume Manager handler for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/

#ifndef __MNTD_VOLUME_MANAGER_H__
#define __MNTD_VOLUME_MANAGER_H__

#include <glib.h>
#include <dbus/dbus.h>

#include "mntd_volume_manager_types.h"
#include "mntd_volume_types.h"
#include "mntd_volume_hash.h"

#if defined(__cplusplus)
extern "C" {
#endif
	

/** Generate new Volume Manager Object
 *
 * @param config                Path to configfile
 * @return                      Pointer to Volume Manager object,
 *                              NULL if error
 */
PVOLUMEMANAGER new_VolumeManager(const char *config);

	
/** main volume manager structure
 *
 */
struct VOLUMEMANAGER_ {
	/* variables */
	PHASHMAP phm;                /**< Pointer to HashMap object */
	char *base;                  /**< Basepath that VolumeManager will be responsible for (read from configfile) */
	char *config;                /**< Path to configfile */
	
	/** Constructor
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @return                      0 if successful, -1 otherwise
	 * @internal
	 */
	int (*init)								(PVOLUMEMANAGER pvm);

	/** Will destroy the VolumeManager object
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 */
	void (*destroy)						(PVOLUMEMANAGER pvm);
	
	/** Rescan HAL devices
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 */
	void (*rescan)							(PVOLUMEMANAGER pvm);
	
	/** Get specific mount point
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @return                      Mount point, or NULL if error
	 */
	char *(*get_mntpnt)					(PVOLUMEMANAGER pvm, const char *udi);
	
	/** Check if given volume exists
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @return                      1 if found, 0 otherwise
	 * @internal
	 */
	int (*contains)						(PVOLUMEMANAGER pvm, const char *udi);

	/** call function for each mounted volume
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param func                  Pointer to callback function
	 * @param userdata              userdata pointer that will be passed to callback function
	 * @internal
	 */
	void (*foreach_mounted)				(PVOLUMEMANAGER pvm,
												int (*func)(void *data, void *userdata), 
												void *userdata);
												
	/** Add volume to manager
	 *
	 * Will be called from mntd_hal, if corresponding signal is sent.
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @internal
	 */
	void (*add_volume)					(PVOLUMEMANAGER pvm, const char *udi);
	
	/** Remove volume from manager
	 *
	 * Will be called from mntd_hal, if corresponding signal is sent.
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @internal
	 */
	void (*remove_volume)				(PVOLUMEMANAGER pvm, const char *udi);
	
	/** Add volume capability to volume
	 *
	 * Will be called from mntd_hal, if corresponding signal is sent.
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @param capability            Pointer to new capability
	 * @internal
	 */
	void (*new_volume_capability)		(PVOLUMEMANAGER pvm, const char *udi, const char *capability);
	
	/** Add volume property
	 *
	 * Will be called from mntd_hal, if corresponding signal is sent.
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @param property              property name of removed volume
	 * @internal
	 */
	void (*add_volume_property)		(PVOLUMEMANAGER pvm, const char *udi, const char *property);
	
	/** Remove volume property
	 *
	 * Will be called from mntd_hal, if corresponding signal is sent.
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @param property              property name of removed volume
	 * @internal
	 */
	void (*remove_volume_property)	(PVOLUMEMANAGER pvm, const char *udi, const char *property);
	
	/** Change volume property
	 *
	 * Will be called from mntd_hal, if corresponding signal is sent.
	 *
	 * @param pvm                   Pointer to Volume Manager Object
	 * @param udi                   udi (key) of object
	 * @param property              property name of removed volume
	 * @internal
	 */
	void (*change_volume_property)	(PVOLUMEMANAGER pvm, const char *udi, const char *property);
	
};


#if defined(__cplusplus)
}
#endif

#endif /* __MNTD_VOLUME_MANAGER_H__ */
