/***************************************************************************
 * CVSID: $Id: mntd_volume_manager_helper.c,v 1.8 2004/05/23 00:29:08 stefanb Exp $
 *
 * mntd_volume_manager_helper.c : Volume Manager handler for MNT daemon
 *
 * Copyright (C) 2004 Stefan Bambach, <stefan@bambach.biz>
 *
 * Licensed under the GNU General Public License 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libhal.h>

#include "mntd_volume_manager_helper.h"
#include "mntd_volume.h"
#include "errmanager.h"


/**
 * @defgroup DaemonVolumeManagerHelper Volume Manager Helper
 * @ingroup MntDaemon
 * @brief Some helper functions
 * @{
 */


/** Check for safe access to device
 *
 * @param udi                   udi of property
 * @return                      1 if safe, 0 otherwise
 */
int
vmh_hal_device_is_safe(const char *udi)
{
	g_assert(udi!=NULL);
	
	// check for device existance
	if (hal_device_exists(udi)) {
		// return OK
		return 1;
	}
	
	// return error
	return 0;
}


/** Check if device is volume
 *
 * @param udi                   udi of property
 * @return                      1 if volume, 0 otherwise
 */
int
vmh_hal_device_is_volume(const char *udi)
{
	char *cat = NULL;
	int res = 0;
	
	g_assert(udi!=NULL);
	
	if (vmh_hal_device_is_safe_property(udi, HAL_PROPERTY_INFO_CATEGORY)) {
		cat = hal_device_get_property_string(udi, HAL_PROPERTY_INFO_CATEGORY);
		if (cat != NULL) {
			if (strcmp(cat, HAL_PROPERTY_INFO_CATEGORY_VOLUME)==0) {
				if (hal_device_get_property_bool(udi, HAL_PROPERTY_BLOCK_IS_VOLUME)) {
					res = 1;
				}
			}
			free(cat);
			cat = NULL;
		}
	}
	
	return res;
}


/** Check for safe access to property
 *
 * @param udi                   udi of property
 * @param key                   key of property
 * @return                      1 if safe, 0 otherwise
 */
int
vmh_hal_device_is_safe_property(const char *udi, const char *key)
{
	g_assert(udi!=NULL);
	g_assert(key!=NULL);
	
	// check for device existance
	if (vmh_hal_device_is_safe(udi)) {
		if (hal_device_property_exists(udi, key)) {
			// return OK
			return 1;
		}
	}
	
	// return error
	return 0;
}


/** Get safe HAL property string
 *
 * @param udi                   udi of property
 * @param key                   key of property
 * @param result                pointer to result
 * @return                      1 if result valid, 0 otherwise
 */
int
vmh_hal_device_get_safe_property_string(const char *udi, const char *key, char **result)
{
	char *hal_string = NULL;
	char *tmp = NULL;

	g_assert(udi!=NULL);
	g_assert(key!=NULL);
	g_assert(result!=NULL);
	
	if (vmh_hal_device_is_safe_property(udi, key)) {
		hal_string = hal_device_get_property_string(udi, key);
		if (hal_string != NULL) {
			tmp = strdup(hal_string);
			if (tmp != NULL) {
				*result = tmp;
				return 1;
			} else {
				*result = NULL;
			}
		}
	}
	
	if (hal_string != NULL) hal_free_string(hal_string);
	if (tmp != NULL) free(tmp);

	// return error
	return 0;
}


/** Get safe HAL property int
 *
 * @param udi                   udi of property
 * @param key                   key of property
 * @param result                pointer to result
 * @return                      1 if result valid, 0 otherwise
 */
int
vmh_hal_device_get_safe_property_int(const char *udi, const char *key, int *result)
{
	g_assert(udi!=NULL);
	g_assert(key!=NULL);
	g_assert(result!=NULL);
	
	if (vmh_hal_device_is_safe_property(udi, key)) {
		*result = hal_device_get_property_int(udi, key);
		return 1;
	}
	
	return 0;
}


/** Get safe HAL property double
 *
 * @param udi                   udi of property
 * @param key                   key of property
 * @param result                pointer to result
 * @return                      1 if result valid, 0 otherwise
 */
int
vmh_hal_device_get_safe_property_double(const char *udi, const char *key, double *result)
{
	g_assert(udi!=NULL);
	g_assert(key!=NULL);
	g_assert(result!=NULL);
	
	if (vmh_hal_device_is_safe_property(udi, key)) {
		*result = hal_device_get_property_double(udi, key);
		return 1;
	}
	
	return 0;
}


/** Get safe HAL property bool
 *
 * @param udi                   udi of property
 * @param key                   key of property
 * @param result                pointer to result
 * @return                      1 if result valid, 0 otherwise
 */
int
vmh_hal_device_get_safe_property_bool(const char *udi, const char *key, int *result)
{
	g_assert(udi!=NULL);
	g_assert(key!=NULL);
	g_assert(result!=NULL);
	
	if (vmh_hal_device_is_safe_property(udi, key)) {
		*result = hal_device_get_property_bool(udi, key);
		return 1;
	}
	
	return 0;
}


/** Helper Set volume device name
 *
 * @param pv                    Pointer to Volume Object
 * @param udi                   udi (key) of object
 * @return                      0 if successful, -1 otherwise
 */
int 
vmh_try_n_set_volumes_device(PVOLUME pv, const char *udi)
{
	char *key = HAL_PROPERTY_BLOCK_DEVICE;
	char *device = NULL;
	int res = 0;

	g_assert(pv!=NULL);
	g_assert(udi!=NULL);

	// getting device name
	if (vmh_hal_device_get_safe_property_string(udi, key, &device)) {
		if ((pv->set_device(pv, device)) == -1) {
			MSG_WARNING("Couldn't set device '%s' for '%s'", device, udi);
			goto error;
		}
	} else {
		if ((pv->set_device(pv, NULL)) == -1) {
			MSG_WARNING("Couldn't remove device for '%s'", udi);
			goto error;
		}
	}
	
	goto cleanup;
	
error:
	res = -1;

cleanup:
	if (device != NULL) {
		free(device);
		device = NULL;
	}
	
	return res;
}


/** Helper Set volume mntpnt
 *
 * @param pv                    Pointer to Volume Object
 * @param udi                   udi (key) of object
 * @return                      0 if successful, -1 otherwise
 */
int 
vmh_try_n_set_volumes_mntpnt(PVOLUME pv, const char *udi)
{
	char *key = HAL_PROPERTY_BLOCK_MOUNT_POINT;
	char *mntpnt = NULL;
	int res = 0;

	g_assert(pv!=NULL);
	g_assert(udi!=NULL);

	// getting device name
	if (vmh_hal_device_get_safe_property_string(udi, key, &mntpnt)) {
		if ((pv->set_mntpnt(pv, mntpnt)) == -1) {
			MSG_WARNING("Couldn't set mntpnt '%s' for '%s'", mntpnt, udi);
			goto error;
		}
	} else {
		if ((pv->set_mntpnt(pv, NULL)) == -1) {
			MSG_WARNING("Couldn't remove mount point for '%s'", udi);
			goto error;
		}
	}
	
	goto cleanup;
	
error:
	res = -1;

cleanup:
	if (mntpnt != NULL) {
		free(mntpnt);
		mntpnt = NULL;
	}
	
	return res;
}


/** Helper Set volume fstype
 *
 * @param pv                    Pointer to Volume Object
 * @param udi                   udi (key) of object
 * @return                      0 if successful, -1 otherwise
 */
int 
vmh_try_n_set_volumes_fstype(PVOLUME pv, const char *udi)
{
	char *key = HAL_PROPERTY_BLOCK_FSTYPE;
	char *fstype = NULL;
	int res = 0;

	g_assert(pv!=NULL);
	g_assert(udi!=NULL);

	// getting device name
	if (vmh_hal_device_get_safe_property_string(udi, key, &fstype)) {
		if ((pv->set_fstype(pv, fstype)) == -1) {
			MSG_WARNING("Couldn't set fstype '%s' for '%s'", fstype, udi);
			goto error;
		}
	} else {
		if ((pv->set_fstype(pv, NULL)) == -1) {
			MSG_WARNING("Couldn't remove filesystem type for '%s'", udi);
			goto error;
		}
	}
	
	goto cleanup;
	
error:
	res = -1;

cleanup:
	if (fstype != NULL) {
		free(fstype);
		fstype = NULL;
	}
	
	return res;
}


/** Helper Set volumes data
 *
 * @param pv                    Pointer to Volume Object
 * @param udi                   udi (key) of object
 * @return                      0 if successful, -1 otherwise
 */
int 
vmh_try_n_set_volumes_data(PVOLUME pv, const char *udi)
{
	g_assert(pv!=NULL);
	g_assert(udi!=NULL);
	
	// setting mount point
	vmh_try_n_set_volumes_mntpnt(pv, udi);
	
	// setting file system type
	vmh_try_n_set_volumes_fstype(pv, udi);
	
	// setting device name
	vmh_try_n_set_volumes_device(pv, udi);
	
	return 0;
}


/** Helper Set volumes data for given property
 *
 * @param pv                    Pointer to Volume Object
 * @param udi                   udi (key) of object
 * @param property              name of property
 * @return                      0 if successful, -1 otherwise
 */
int 
vmh_try_n_set_volumes_data_by_property(PVOLUME pv, const char *udi, const char *property)
{
	g_assert(pv!=NULL);
	g_assert(udi!=NULL);
	g_assert(property!=NULL);
	
	// setting mount point
	if (strcmp(property, HAL_PROPERTY_BLOCK_MOUNT_POINT)==0) {
		vmh_try_n_set_volumes_mntpnt(pv, udi);
	}
	
	// setting file system type
	if (strcmp(property, HAL_PROPERTY_BLOCK_FSTYPE)==0) {
		vmh_try_n_set_volumes_fstype(pv, udi);
	}
				
	// setting device name
	if (strcmp(property, HAL_PROPERTY_BLOCK_DEVICE)==0) {
		vmh_try_n_set_volumes_device(pv, udi);
	}
	
	return 0;
}


/** @} */
