/**
 * PThreads
 *
 * @defgroup thread Thread implementation
 * @author Stefan Bambach (sbambach@gmx.net)
 * @version 0.1
 * @{
 */

#ifndef __HELPER_THREADING_H__
#define __HELPER_THREADING_H__

#ifdef HAVE_CONFIG_H
	#include "config.h"
#endif

#ifdef WITH_PTHREADS
   #include <sys/time.h>
   #include <pthread.h>
   #include <errno.h>
   extern int errno;
#elif defined(WITH_SDLTHREADS)
   #include <SDL_thread.h>
#endif
#include <stdlib.h>
#include "errmanager.h"


/*
 * public defines
 */


typedef void * thread_rdwrattr_t;					/**< read/write lock pointer */

#define thread_rdwrattr_default NULL;           /**< default value */
#define T_RDWR thread_rdwr_t                    /**< wrapper for read/write thread struct */
#define T_RDWRATTR thread_rdwrattr_t            /**< wrapper for read/write attr struct */

#ifdef WITH_PTHREADS
   #define T_MUTEX pthread_mutex_t              /**< wrapper for mutex */
   #define T_MUTEXATTR pthread_mutexattr_t      /**< wrapper for mutex attributes */
   #define T_COND pthread_cond_t                /**< wrapper for conditional */
   #define T_CONDATTR pthread_condattr_t        /**< wrapper for conditional attributes */
   #define T_THREAD pthread_t                   /**< wrapper for thread */
   #define T_THREADATTR pthread_attr_t          /**< wrapper for thread attributes */
#elif defined(WITH_SDLTHREADS)
   #define T_MUTEX SDL_mutex *                  /**< wrapper for mutex */
   #define T_MUTEXATTR int                      /**< wrapper for mutex attributes */
   #define T_COND SDL_cond *                    /**< wrapper for conditional */
   #define T_CONDATTR int                       /**< wrapper for conditional attributes */
   #define T_THREAD SDL_Thread *                /**< wrapper for thread */
   #define T_THREADATTR int                     /**< wrapper for thread attributes */
#else
   #define T_MUTEX int                          /**< dummy wrapper for mutex */
   #define T_MUTEXATTR int                      /**< dummy wrapper for mutex attributes */
   #define T_COND int                           /**< dummy wrapper for conditional */
   #define T_CONDATTR int                       /**< dummy wrapper for conditional attributes */
   #define T_THREAD int                         /**< dummy wrapper for thread */
   #define T_THREADATTR int                     /**< dummy wrapper for thread attributes */
#endif

#define T_MUTEX_INIT(mutex, attr) thread_mutex_init(mutex, attr)           /**< wrapper for initialize mutex lock */
#define T_MUTEX_LOCK(mutex) thread_mutex_lock(mutex)                       /**< wrapper for setting mutex lock */
#define T_MUTEX_TRYLOCK(mutex) thread_mutex_trylock(mutex)                 /**< wrapper for setting mutex try lock */
#define T_MUTEX_UNLOCK(mutex) thread_mutex_unlock(mutex)                   /**< wrapper for releasing mutex lock */
#define T_MUTEX_DESTROY(mutex) thread_mutex_destroy(mutex)                 /**< wrapper for destroying mutex */
#define T_COND_INIT(cond, attr) thread_cond_init(cond, attr);              /**< wrapper for initialize conditional */
#define T_COND_SIGNAL(cond) thread_cond_signal(cond)                       /**< wrapper for sending conditional signal */
#define T_COND_BROADCAST(cond) thread_cond_broadcast(cond)                 /**< wrapper for sending conditional broadcast */
#define T_COND_WAIT(cond, mutex) thread_cond_wait(cond, mutex)             /**< wrapper for waiting for conditional wait on mutex */
#define T_COND_TIMEDWAIT(c, m, t) thread_cond_timedwait(c, m, t)           /**< wrapper for waiting for conditional wait on mutex with timeout */
#define T_COND_DESTROY(cond) thread_cond_destroy(cond)                     /**< wrapper for destroying conditional */
#define T_RDWR_INIT(rdwrp, attrp) thread_rdwr_init_np(rdwrp, attrp)        /**< wrapper for initialize read/write lock */
#define T_RDWR_RLOCK(rdwrp) thread_rdwr_rlock_np(rdwrp)                    /**< wrapper for setting read lock */
#define T_RDWR_RUNLOCK(rdwrp) thread_rdwr_runlock_np(rdwrp)                /**< wrapper for releasing read lock */
#define T_RDWR_WLOCK(rdwrp) thread_rdwr_wlock_np(rdwrp)                    /**< wrapper for setting write lock */
#define T_RDWR_WUNLOCK(rdwrp) thread_rdwr_wunlock_np(rdwrp)                /**< wrapper for releasing write lock */
#define T_CREATE(t, attr, func, arg) thread_create(t, attr, func, arg)     /**< wrapper for creating new thread */
#define T_JOIN(t, ret) thread_join(t, ret)                                 /**< wrapper for joining a thread */
#define T_EXIT(ret) thread_exit(ret)                                       /**< wrapper for exiting a thread */
#define T_CANCEL(t) thread_cancel(t)                                       /**< wrapper for cancelling a thread */


/*
 * public structs
 */


/** Reader/Writer lock struct
 *
 */
typedef struct rdwr_var {
	int readers_reading;          /**< Counts rlocks */
	int writer_writing;           /**< Counts wlocks */
	T_MUTEX mutex;                /**< locks the struct itself */
	T_COND lock_free;             /**< signals that the lock is free */
} T_RDWR;


/*
 * public methods
 */


/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////    T H R E A D
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////


/**
 * Create a thread
 *
 * @param thread     Pointer to T_THREAD
 * @param attr       Pointer to T_THREADATTR
 * @param routine    Pointer to thread function
 * @param arg        Pointer to argument for function
 * @return           0 if successful, -1 otherwise.
 */
int thread_create(T_THREAD *thread, T_THREADATTR *attr, void *(*routine)(void *), void *arg);


/**
 * Join a thread
 *
 * @param thread     Pointer to T_THREAD
 * @param ret        Pointer to return value
 * @return           0 if successful, -1 otherwise.
 */
int thread_join(T_THREAD thread, void **ret);


/**
 * Exit a thread
 *
 * @param ret        Pointer to return value
 * @return           0 if successful, -1 otherwise.
 */
int thread_exit(void *ret);


/**
 * Cancel a thread
 *
 * @param thread     Pointer to T_THREAD
 * @return           0 if successful, -1 otherwise.
 */
int thread_cancel(T_THREAD thread);


/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////    C O N D I T I O N A L
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////


/**
 * Initializing a conditional variable
 *
 * @param cond       Pointer to T_COND
 * @param cond_attr  Pointer to T_CONDATTR
 * @return           0 if successful, -1 otherwise.
 */
int thread_cond_init(T_COND *cond, T_CONDATTR *cond_attr);


/**
 * Sending a conditional signal
 *
 * @param cond       Pointer to T_COND
 * @return           0 if successful, -1 otherwise.
 */
int thread_cond_signal(T_COND *cond);


/**
 * Sending a conditional broadcast
 *
 * @param cond       Pointer to T_COND
 * @return           0 if successful, -1 otherwise.
 */
int thread_cond_broadcast(T_COND *cond);


/**
 * Waiting for conditional wait on mutex
 *
 * @param cond       Pointer to T_COND
 * @param mutex      Pointer to T_MUTEX
 * @return           0 if successful, -1 otherwise.
 */
int thread_cond_wait(T_COND *cond, T_MUTEX *mutex);


/**
 * Waiting for conditional wait on mutex with timeout
 *
 * @param cond       Pointer to T_COND
 * @param mutex      Pointer to T_MUTEX
 * @param timeout    Timeout in millisecond
 * @return           0 if successful, -1 if error and 1 if timed out
 */
int thread_cond_timedwait(T_COND *cond, T_MUTEX *mutex, unsigned int timeout);


/**
 * Destroying conditional
 *
 * @param cond       Pointer to T_COND
 * @return           0 if successful, -1 otherwise.
 */
int thread_cond_destroy(T_COND *cond);


/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////    M U T E X
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////


/**
 * Initializing a mutex
 *
 * @param mutex      Pointer to T_MUTEX
 * @param mutexattr  Pointer to T_MUTEXATTR
 * @return           0 if successful, -1 otherwise.
 */
int thread_mutex_init(T_MUTEX *mutex, T_MUTEXATTR *mutexattr);


/**
 * Lock a mutex
 *
 * @param mutex      Pointer to T_MUTEX
 * @return           0 if successful, -1 otherwise.
 */
int thread_mutex_lock(T_MUTEX *mutex);


/**
 * Try to lock a mutex
 *
 * @param mutex      Pointer to T_MUTEX
 * @return           0 if successful, -1 otherwise.
 */
int thread_mutex_trylock(T_MUTEX *mutex);


/**
 * Unlock a mutex
 *
 * @param mutex      Pointer to T_MUTEX
 * @return           0 if successful, -1 otherwise.
 */
int thread_mutex_unlock(T_MUTEX *mutex);


/**
 * Lock a mutex
 *
 * @param mutex      Pointer to T_MUTEX
 * @return           0 if successful, -1 otherwise.
 */
int thread_mutex_destroy(T_MUTEX *mutex);


/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
/////    R E A D E R  /  W R I T E R
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////


/**
 * Initializing a Reader/Writer Lock
 *
 * @param rdwrp      Pointer to T_RDWR struct
 * @param attrp      Pointer to T_RDWRATTR struct
 * @return           0 if successful, -1 otherwise.
 */
int thread_rdwr_init_np(T_RDWR *rdwrp, T_RDWRATTR *attrp);


/**
 * Read Locking a Reader/Writer Lock
 *
 * @param rdwrp      Pointer to T_RDWR struct
 * @return           0 if successful, -1 otherwise.
 */
int thread_rdwr_rlock_np(T_RDWR *rdwrp);


/**
 * Read Unlocking a Reader/Writer Lock
 *
 * @param rdwrp      Pointer to T_RDWR struct
 * @return           0 if successful, -1 otherwise.
 */
int thread_rdwr_runlock_np(T_RDWR *rdwrp);


/**
 * Write Locking a Reader/Writer Lock
 *
 * @param rdwrp      Pointer to T_RDWR struct
 * @return           0 if successful, -1 otherwise.
 */
int thread_rdwr_wlock_np(T_RDWR *rdwrp);


/**
 * Write Unlocking a Reader/Writer Lock
 *
 * @param rdwrp      Pointer to T_RDWR struct
 * @return           0 if successful, -1 otherwise.
 */
int thread_rdwr_wunlock_np(T_RDWR *rdwrp);


/*
 * private methods
 */




#endif

/** @} */
